#!/usr/bin/env python3
"""
Multiple aligns sequence fields
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import shutil
from argparse import ArgumentParser
from collections import OrderedDict
from itertools import chain
from textwrap import dedent
from Bio.SeqRecord import SeqRecord

# Presto and changeo import
from presto.Defaults import default_out_args, default_muscle_exec
from presto.Applications import runMuscle
from presto.IO import printLog, printError, printWarning
from presto.Multiprocessing import manageProcesses
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from changeo.IO import getDbFields, getFormatOperators
from changeo.Multiprocessing import DbResult, feedDbQueue, processDbQueue, collectDbQueue


# TODO:  maybe not bothering with 'set' is best. can just work off field identity
def groupRecords(records, fields=None, calls=['v', 'j'], mode='gene', action='first'):
    """
    Groups Receptor objects based on gene or annotation

    Arguments:
      records : an iterator of Receptor objects to group.
      fields : gene field to group by.
      calls : allele calls to use for grouping.
              one or more of ('v', 'd', 'j').
      mode : specificity of alignment call to use for allele call fields.
             one of ('allele', 'gene').
      action : only 'first' is currently supported.

    Returns:
    dictionary of grouped records
    """
    # Define functions for grouping keys
    if mode == 'allele' and fields is None:
        def _get_key(rec, calls, action):
            return tuple(rec.getAlleleCalls(calls, action))
    elif mode == 'gene' and fields is None:
        def _get_key(rec, calls, action):
            return tuple(rec.getGeneCalls(calls, action))
    elif mode == 'allele' and fields is not None:
        def _get_key(rec, calls, action):
            vdj = rec.getAlleleCalls(calls, action)
            ann = [rec.getChangeo(k) for k in fields]
            return tuple(chain(vdj, ann))
    elif mode == 'gene' and fields is not None:
        def _get_key(rec, calls, action):
            vdj = rec.getGeneCalls(calls, action)
            ann = [rec.getChangeo(k) for k in fields]
            return tuple(chain(vdj, ann))

    rec_index = {}
    for rec in records:
        key = _get_key(rec, calls, action)
        # Assigned grouped records to individual keys and all failed to a single key
        if all([k is not None for k in key]):
            rec_index.setdefault(key, []).append(rec)
        else:
            rec_index.setdefault(None, []).append(rec)

    return rec_index


def alignBlocks(data, field_map, muscle_exec=default_muscle_exec):
    """
    Multiple aligns blocks of sequence fields together

    Arguments:
      data : DbData object with Receptor objects to process.
      field_map : a dictionary of {input sequence : output sequence) field names to multiple align.
      muscle_exec : the MUSCLE executable.

    Returns:
      changeo.Multiprocessing.DbResult : object containing Receptor objects with multiple aligned sequence fields.
    """
    # Define sequence fields
    seq_fields = list(field_map.keys())

    # Function to validate record
    def _pass(rec):
        if all([len(rec.getField(f)) > 0 for f in seq_fields]):
            return True
        else:
            return False

    # Define return object
    result = DbResult(data.id, data.data)
    result.results = data.data
    result.valid = True

    # Fail invalid groups
    if result.id is None or not all([_pass(x) for x in data.data]):
        result.log = None
        result.valid = False
        return result

    # Run muscle and map results
    seq_list = [SeqRecord(r.getSeq(f), id='%s_%s' % (r.sequence_id.replace(' ', '_'), f)) for f in seq_fields \
                for r in data.data]
    seq_aln = runMuscle(seq_list, aligner_exec=muscle_exec)
    if seq_aln is not None:
        aln_map = {x.id: i for i, x in enumerate(seq_aln)}
        for i, r in enumerate(result.results, start=1):
            for f in seq_fields:
                idx = aln_map['%s_%s' % (r.sequence_id.replace(' ', '_'), f)]
                seq = str(seq_aln[idx].seq)
                r.annotations[field_map[f]] = seq
                result.log['%s-%s' % (f, r.sequence_id)] = seq

    else:
        result.valid = False

    #for r in result.results:  print r.annotations
    return result


def alignAcross(data, field_map, muscle_exec=default_muscle_exec):
    """
    Multiple aligns sequence fields column wise

    Arguments:
      data : DbData object with Receptor objects to process.
      field_map : a dictionary of {input sequence : output sequence) field names to multiple align.
      muscle_exec : the MUSCLE executable.

    Returns:
      changeo.Multiprocessing.DbResult : object containing Receptor objects with multiple aligned sequence fields.
    """
    # Define sequence fields
    seq_fields = list(field_map.keys())

    # Function to validate record
    def _pass(rec):
        if all([len(rec.getField(f)) > 0 for f in seq_fields]):
            return True
        else:
            return False

    # Define return object
    result = DbResult(data.id, data.data)
    result.results = data.data
    result.valid = True

    # Fail invalid groups
    if result.id is None or not all([_pass(x) for x in data.data]):
        result.log = None
        result.valid = False
        return result

    seq_fields = list(field_map.keys())
    for f in seq_fields:
        seq_list = [SeqRecord(r.getSeq(f), id=r.sequence_id.replace(' ', '_')) for r in data.data]
        seq_aln = runMuscle(seq_list, aligner_exec=muscle_exec)
        if seq_aln is not None:
            aln_map = {x.id: i for i, x in enumerate(seq_aln)}
            for i, r in enumerate(result.results, start=1):
                idx = aln_map[r.sequence_id.replace(' ', '_')]
                seq = str(seq_aln[idx].seq)
                r.annotations[field_map[f]] = seq
                result.log['%s-%s' % (f, r.sequence_id)] = seq
        else:
            result.valid = False

    #for r in result.results:  print r.annotations
    return result


def alignWithin(data, field_map, muscle_exec=default_muscle_exec):
    """
    Multiple aligns sequence fields within a row

    Arguments:
      data : DbData object with Receptor objects to process.
      field_map : a dictionary of {input sequence : output sequence) field names to multiple align.
      muscle_exec : the MUSCLE executable.

    Returns:
      changeo.Multiprocessing.DbResult : object containing Receptor objects with multiple aligned sequence fields.
    """
    # Define sequence fields
    seq_fields = list(field_map.keys())

    # Function to validate record
    def _pass(rec):
        if all([len(rec.getField(f)) > 0 for f in seq_fields]):
            return True
        else:
            return False

    # Define return object
    result = DbResult(data.id, data.data)
    result.results = data.data
    result.valid = True

    # Fail invalid groups
    if result.id is None or not _pass(data.data):
        result.log = None
        result.valid = False
        return result

    record = data.data
    seq_list = [SeqRecord(record.getSeq(f), id=f) for f in seq_fields]
    seq_aln = runMuscle(seq_list, aligner_exec=muscle_exec)
    if seq_aln is not None:
        aln_map = {x.id: i for i, x in enumerate(seq_aln)}
        for f in seq_fields:
            idx = aln_map[f]
            seq = str(seq_aln[idx].seq)
            record.annotations[field_map[f]] = seq
            result.log[f] = seq
    else:
        result.valid = False

    return result


def alignRecords(db_file, seq_fields, group_func, align_func, group_args={}, align_args={},
                 format='changeo', out_file=None, out_args=default_out_args, nproc=None, queue_size=None):
    """
    Performs a multiple alignment on sets of sequences

    Arguments: 
      db_file : filename of the input database.
      seq_fields : the sequence fields to multiple align.
      group_func : function to use to group records.
      align_func : function to use to multiple align sequence groups.
      group_args : dictionary of arguments to pass to group_func.
      align_args : dictionary of arguments to pass to align_func.
      format : output format. One of 'changeo' or 'airr'.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
      nproc : the number of processQueue processes.
              if None defaults to the number of CPUs.
      queue_size : maximum size of the argument queue.
                   if None defaults to 2*nproc.
                      
    Returns: 
      dict : names of the 'pass' and 'fail' output files.
    """
    # Define subcommand label dictionary
    cmd_dict = {alignAcross: 'across', alignWithin: 'within', alignBlocks: 'block'}
    
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'AlignRecords'
    log['COMMAND'] = cmd_dict.get(align_func, align_func.__name__)
    log['FILE'] = os.path.basename(db_file)
    log['SEQ_FIELDS'] = ','.join(seq_fields)
    if 'group_fields' in group_args: log['GROUP_FIELDS'] = ','.join(group_args['group_fields'])
    if 'mode' in group_args: log['MODE'] = group_args['mode']
    if 'action' in group_args: log['ACTION'] = group_args['action']
    log['NPROC'] = nproc
    printLog(log)

    # Define format operators
    try:
        reader, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)

    # Define feeder function and arguments
    if 'group_fields' in group_args and group_args['group_fields'] is not None:
        group_args['group_fields'] = [schema.toReceptor(f) for f in group_args['group_fields']]
    feed_func = feedDbQueue
    feed_args = {'db_file': db_file,
                 'reader': reader,
                 'group_func': group_func,
                 'group_args': group_args}
    # Define worker function and arguments
    field_map = OrderedDict([(schema.toReceptor(f), '%s_align' % f) for f in seq_fields])
    align_args['field_map'] = field_map
    work_func = processDbQueue
    work_args = {'process_func': align_func,
                 'process_args': align_args}
    # Define collector function and arguments
    out_fields = getDbFields(db_file, add=list(field_map.values()), reader=reader)
    out_args['out_type'] = schema.out_type
    collect_func = collectDbQueue
    collect_args = {'db_file': db_file,
                    'label': 'align',
                    'fields': out_fields,
                    'writer': writer,
                    'out_file': out_file,
                    'out_args': out_args}
    
    # Call process manager
    result = manageProcesses(feed_func, work_func, collect_func, 
                             feed_args, work_args, collect_args, 
                             nproc, queue_size)
        
    # Print log
    result['log']['END'] = 'AlignRecords'
    printLog(result['log'])
    output = {k: v for k, v in result.items() if k in ('pass', 'fail')}

    return output


def getArgParser():
    """
    Defines the ArgumentParser

    Arguments: 
    None
                      
    Returns: 
    an ArgumentParser object
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 align-pass
                     database with multiple aligned sequences.
                 align-fail
                     database with records failing alignment.

             required fields:
                 sequence_id, v_call, j_call
                 <field>
                     user specified sequence fields to align.

             output fields:
                 <field>_align
             ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', dest='command', metavar='',
                                       help='alignment method')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser
    parser_parent = getCommonArgParser(format=True, multiproc=True)

    # Argument parser for column-wise alignment across records
    parser_across = subparsers.add_parser('across', parents=[parser_parent],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='''Multiple aligns sequence columns within groups 
                                                 and across rows using MUSCLE.''')
    group_across = parser_across.add_argument_group('alignment arguments')
    group_across.add_argument('--sf', nargs='+', action='store', dest='seq_fields', required=True,
                               help='The sequence fields to multiple align within each group.')
    group_across.add_argument('--gf', nargs='+', action='store', dest='group_fields', default=None,
                               help='Additional (not allele call) fields to use for grouping.')
    group_across.add_argument('--calls', nargs='+', action='store', dest='calls',
                               choices=('v', 'd', 'j'), default=['v', 'j'],
                               help='Segment calls (allele assignments) to use for grouping.')
    group_across.add_argument('--mode', action='store', dest='mode',
                              choices=('allele', 'gene'), default='gene',
                              help='''Specifies whether to use the V(D)J allele or gene when
                                   an allele call field (--calls) is specified.''')
    group_across.add_argument('--act', action='store', dest='action', default='first',
                               choices=('first', ),
                               help='''Specifies how to handle multiple values within default
                                     allele call fields. Currently, only "first" is supported.''')
    group_across.add_argument('--exec', action='store', dest='muscle_exec',
                               default=default_muscle_exec,
                               help='The location of the MUSCLE executable')
    parser_across.set_defaults(group_func=groupRecords, align_func=alignAcross)


    # Argument parser for alignment of fields within records
    parser_within = subparsers.add_parser('within', parents=[parser_parent],
                                          formatter_class=CommonHelpFormatter, add_help=False,
                                          help='Multiple aligns sequence fields within rows using MUSCLE')
    group_within = parser_within.add_argument_group('alignment arguments')
    group_within.add_argument('--sf', nargs='+', action='store', dest='seq_fields', required=True,
                               help='The sequence fields to multiple align within each record.')
    group_within.add_argument('--exec', action='store', dest='muscle_exec',
                              default=default_muscle_exec,
                              help='The location of the MUSCLE executable')
    parser_within.set_defaults(group_func=None, align_func=alignWithin)

    # Argument parser for column-wise alignment across records
    parser_block = subparsers.add_parser('block', parents=[parser_parent],
                                        formatter_class=CommonHelpFormatter, add_help=False,
                                        help='''Multiple aligns sequence groups across both 
                                             columns and rows using MUSCLE.''')
    group_block = parser_block.add_argument_group('alignment arguments')
    group_block.add_argument('--sf', nargs='+', action='store', dest='seq_fields', required=True,
                               help='The sequence fields to multiple align within each group.')
    group_block.add_argument('--gf', nargs='+', action='store', dest='group_fields', default=None,
                               help='Additional (not allele call) fields to use for grouping.')
    group_block.add_argument('--calls', nargs='+', action='store', dest='calls',
                               choices=('v', 'd', 'j'), default=['v', 'j'],
                               help='Segment calls (allele assignments) to use for grouping.')
    group_block.add_argument('--mode', action='store', dest='mode',
                              choices=('allele', 'gene'), default='gene',
                              help='''Specifies whether to use the V(D)J allele or gene when
                                   an allele call field (--calls) is specified.''')
    group_block.add_argument('--act', action='store', dest='action', default='first',
                               choices=('first', ),
                               help='''Specifies how to handle multiple values within default
                                     allele call fields. Currently, only "first" is supported.''')
    group_block.add_argument('--exec', action='store', dest='muscle_exec',
                               default=default_muscle_exec,
                               help='The location of the MUSCLE executable')
    parser_block.set_defaults(group_func=groupRecords, align_func=alignBlocks)

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Check if a valid MUSCLE executable was specified for muscle mode
    if not shutil.which(args.muscle_exec):
        parser.error('%s does not exist or is not executable.' % args.muscle_exec)
    # Define align_args
    args_dict['align_args'] = {'muscle_exec': args_dict['muscle_exec']}
    del args_dict['muscle_exec']

    # Define group_args
    if args_dict['group_func'] is groupRecords:
        args_dict['group_args'] = {'fields':args_dict['group_fields'],
                                   'calls':args_dict['calls'],
                                   'mode':args_dict['mode'],
                                   'action':args_dict['action']}
        del args_dict['group_fields']
        del args_dict['calls']
        del args_dict['mode']
        del args_dict['action']

    # Clean arguments dictionary
    del args_dict['command']
    del args_dict['db_files']
    if 'out_files' in args_dict: del args_dict['out_files']

    # Call main function for each input file
    for i, f in enumerate(args.__dict__['db_files']):
        args_dict['db_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        alignRecords(**args_dict)




"""
File I/O and parsers
"""
# Info
__author__ = 'Namita Gupta, Jason Anthony Vander Heiden'

# Imports
import csv
import os
import re
import tarfile
import yaml
import zipfile
from itertools import chain, groupby, zip_longest
from tempfile import TemporaryDirectory
from textwrap import indent
from Bio import SeqIO
from Bio.Seq import Seq

# Presto and changeo imports
from presto.IO import getFileType, printError, printWarning, printDebug
from changeo.Defaults import default_csv_size
from changeo.Gene import getAllele, getLocus, getVAllele, getDAllele, getJAllele, getCAllele
from changeo.Receptor import AIRRSchema, AIRRSchemaAA, ChangeoSchema, ChangeoSchemaAA, Receptor, ReceptorData
from changeo.Alignment import decodeBTOP, encodeCIGAR, padAlignment, gapV, inferJunction, \
                              RegionDefinition, getRegions

# System settings
csv.field_size_limit(default_csv_size)


class TSVReader:
    """
    Simple csv.DictReader wrapper to read format agnostic TSV files.

    Attributes:
      reader (iter): reader object.
      fields (list): field names.
    """

    def __init__(self, handle):
        """
        Initializer

        Arguments:
          handle : handle to an open TSV file

        Returns:
          changeo.IO.TSVReader
        """
        # Arguments
        self.handle = handle
        self.receptor = False
        self.reader = csv.DictReader(self.handle, dialect='excel-tab')
        self.fields = self.reader.fieldnames

    def __iter__(self):
        """
        Iterator initializer

        Returns:
          changeo.IO.TSVReader
        """
        return self

    def __next__(self):
        """
        Next method

        Returns:
          dist : row as a dictionary of field:value pairs.
        """
        # Get next row from reader iterator
        try:
            record = next(self.reader)
        except StopIteration:
            self.handle.close()
            raise StopIteration

        return self._parse(record)

    def _parse(self, record):
        """
        Parses a dictionary of fields

        Arguments:
          record : dict with fields and values to parse.

        Returns:
          dict : parsed dict.
        """
        return record


class TSVWriter:
    """
    Simple csv.DictWriter wrapper to write format agnostic TSV files.
    """
    def __init__(self, handle, fields, header=True):
        """
        Initializer

        Arguments:
          handle : handle to an open output file
          fields : list of output field names
          header : if True write the header on initialization.

        Returns:
          changeo.IO.TSVWriter
        """
        # Arguments
        self.handle = handle
        self.fields = fields
        self.writer = csv.DictWriter(self.handle, fieldnames=self.fields,
                                     dialect='excel-tab', extrasaction='ignore',
                                     lineterminator='\n')
        if header:
            self.writeHeader()

    def writeHeader(self):
        """
        Writes the header

        Returns:
          None
        """
        self.writer.writeheader()

    def writeDict(self, records):
        """
        Writes a row from a dictionary

        Arguments:
          records : dictionary of row data or an iterable of such objects.

        Returns:
          None
        """
        if isinstance(records, dict):
            self.writer.writerow(records)
        else:
            self.writer.writerows(records)


class ChangeoReader(TSVReader):
    """
    An iterator to read and parse Change-O formatted data.
    """
    def __init__(self, handle):
        """
        Initializer

        Arguments:
          handle : handle to an open Change-O formatted file

        Returns:
          changeo.IO.ChangeoReader
        """
        # Arguments
        self.handle = handle
        self.reader = csv.DictReader(self.handle, dialect='excel-tab')
        self.reader.fieldnames = [n.strip().upper() for n in self.reader.fieldnames]
        self.fields = self.reader.fieldnames

    def _parse(self, record):
        """
        Parses a dictionary to a Receptor object

        Arguments:
          record : dict with fields and values in the Change-O format

        Returns:
          changeo.Receptor.Receptor : parsed Receptor object.
        """
        # Parse fields
        result = {}
        for k, v in record.items():
            k = ChangeoSchema.toReceptor(k)
            result[k] = v

        return Receptor(result)


class ChangeoWriter(TSVWriter):
    """
    Writes Change-O formatted data.
    """
    def __init__(self, handle, fields=ChangeoSchema.required, header=True):
        """
        Initializer

        Arguments:
          handle : handle to an open output file
          fields : list of output field names
          header : if True write the header on initialization.

        Returns:
          changeo.IO.ChangeoWriter
        """
        # Arguments
        self.handle = handle
        self.fields = [n.strip().upper() for n in fields]
        self.writer = csv.DictWriter(self.handle, fieldnames=self.fields,
                                     dialect='excel-tab', extrasaction='ignore',
                                     lineterminator='\n')
        if header:
            self.writeHeader()

    def _parseReceptor(self, record):
        """
        Parses a Receptor object to a Change-O dictionary

        Arguments:
          record : dict with fields and values in the Receptor format.

        Returns:
          dict : parsed dict.
        """
        row = record.toDict()
        # Parse known fields
        result = {}
        for k, v in row.items():
            k = ChangeoSchema.fromReceptor(k)
            result[k] = v

        return result

    def writeReceptor(self, records):
        """
        Writes a row from a Receptor object

        Arguments:
          records : a changeo.Receptor.Receptor object to write or an iterable of such objects.

        Returns:
          None
        """
        if isinstance(records, Receptor):
            row = self._parseReceptor(records)
            self.writer.writerow(row)
        else:
            rows = (self._parseReceptor(r) for r in records)
            self.writer.writerows(rows)


class AIRRReader(TSVReader):
    """
    An iterator to read and parse AIRR formatted data.
    """
    def __init__(self, handle):
        """
        Initializer

        Arguments:
          handle : handle to an open AIRR formatted file
          receptor : if True (default) iteration returns a Receptor object, otherwise it returns a dictionary.

        Returns:
          changeo.IO.AIRRReader
        """
        # Arguments
        self.handle = handle

        # Define reader
        try:
            import airr
            self.reader = airr.io.RearrangementReader(self.handle, base=0, debug=False)
        except ImportError as e:
            printError('AIRR library cannot be imported: %s.' % e)

        # Set field list
        self.fields = self.reader.fields

    def _parse(self, record):
        """
        Parses a dictionary of AIRR records to a Receptor object

        Arguments:
          record : dict with fields and values in the AIRR format.

        Returns:
          changeo.Receptor.Receptor : parsed Receptor object.
        """
        # Parse fields
        result = {}
        for k, v in record.items():
            # Rename fields
            k = AIRRSchema.toReceptor(k)
            # Convert start positions to 0-based
            # if k in ReceptorData.start_fields and v is not None and v != '':
            #     v = str(int(v) + 1)
            # Assign new field
            result[k] = v

        # Assign length based on start and end
        for end, (start, length) in ReceptorData.end_fields.items():
            if end in result and result[end] is not None:
                result[length] = int(result[end]) - int(result[start]) + 1

        return Receptor(result)


class AIRRWriter(TSVWriter):
    """
    Writes AIRR formatted data.
    """
    def __init__(self, handle, fields=AIRRSchema.required):
        """
        Initializer

        Arguments:
          handle : handle to an open output file
          fields : list of output field names

        Returns:
          changeo.IO.AIRRWriter
        """
        # Arguments
        self.handle = handle
        self.fields = [f.lower() for f in fields]

        # Define writer
        try:
            import airr
            self.writer = airr.io.RearrangementWriter(self.handle, fields=self.fields,
                                                      base=0, debug=False)
        except ImportError as e:
            printError('AIRR library cannot be imported: %s.' % e)

    def _parseReceptor(self, record):
        """
        Parses a Receptor object to an AIRR dictionary

        Arguments:
          record : dict with fields and values in the Receptor format

        Returns:
          dict : a parsed dict.
        """
        result = {}
        row = record.toDict()
        for k, v in row.items():
            # Convert start positions to 0-based
            # if k in ReceptorData.start_fields and v is not None and v != '':
            #     v = str(int(v) - 1)
            # Convert field names
            k = AIRRSchema.fromReceptor(k)
            result[k] = v

        return result

    def writeReceptor(self, records):
        """
        Writes a row from a Receptor object

        Arguments:
          records : a changeo.Receptor object to write or iterable of such objects.

        Returns:
          None
        """
        if isinstance(records, Receptor):
            row = self._parseReceptor(records)
            self.writer.write(row)
        else:
            rows = (self._parseReceptor(r) for r in records)
            for r in rows:  self.writer.write(r)


class IMGTReader:
    """
    An iterator to read and parse IMGT output files.
    """
    @staticmethod
    def customFields(scores=False, regions=False, junction=False, schema=None):
        """
        Returns non-standard fields defined by the parser

        Arguments:
          scores : if True include alignment scoring fields.
          regions : if True include IMGT-gapped CDR and FWR region fields.
          junction : if True include detailed junction annotation fields.
          schema : schema class to pass field through for conversion.
                   If None, return changeo.Receptor.Receptor attribute names.

        Returns:
          list : list of field names.
        """
        # Alignment scoring fields
        score_fields = ['v_score',
                        'v_identity',
                        'j_score',
                        'j_identity']

        # FWR amd CDR fields
        region_fields = ['fwr1_imgt',
                         'fwr2_imgt',
                         'fwr3_imgt',
                         'fwr4_imgt',
                         'cdr1_imgt',
                         'cdr2_imgt',
                         'cdr3_imgt']

        # Define default detailed junction field ordering
        junction_fields = ['n1_length',
                           'n2_length',
                           'p3v_length',
                           'p5d_length',
                           'p3d_length',
                           'p5j_length',
                           'd_frame']


        fields = []
        if scores:  fields.extend(score_fields)
        if regions:  fields.extend(region_fields)
        if junction:  fields.extend(junction_fields)

        # Convert field names if schema provided
        if schema is not None:
            fields = [schema.fromReceptor(f) for f in fields]

        return fields

    def __init__(self, summary, gapped, ntseq, junction, receptor=True):
        """
        Initializer

        Arguments:
          summary : handle to an open '1_Summary' IMGT/HighV-QUEST output file.
          gapped : handle to an open '2_IMGT-gapped-nt-sequences' IMGT/HighV-QUEST output file.
          ntseq: handle to an open '3_Nt-sequences' IMGT/HighV-QUEST output file.
          junction : handle to an open '6_Junction' IMGT/HighV-QUEST output file.
          receptor : if True (default) iteration returns an Receptor object, otherwise it returns a dictionary.

        Returns:
          change.Parsers.IMGTReader
        """
        # Arguments
        self.summary = summary
        self.gapped = gapped
        self.ntseq = ntseq
        self.junction = junction
        self.receptor = receptor

        # Open readers
        readers = [csv.DictReader(self.summary, delimiter='\t'),
                   csv.DictReader(self.gapped, delimiter='\t'),
                   csv.DictReader(self.ntseq, delimiter='\t'),
                   csv.DictReader(self.junction, delimiter='\t')]
        self.records = zip(*readers)

    @staticmethod
    def _parseFunctionality(summary):
        """
        Parse functionality information

        Arguments:
          summary : dictionary containing one row of the '1_Summary' file.

        Returns:
          dict : database entries for functionality information.
        """
        # Correct for new functionality column names
        if 'Functionality' not in summary:
            summary['Functionality'] = summary['V-DOMAIN Functionality']
            summary['Functionality comment'] = summary['V-DOMAIN Functionality comment']

        # Orientation parser
        def _revcomp():
            x = {'+': 'F', '-': 'T'}
            return x.get(summary['Orientation'], None)

        # Functionality parser
        def _functional():
            x = summary['Functionality']
            if x.startswith('productive'):
                return 'T'
            elif x.startswith('unproductive'):
                return 'F'
            else:
                return None

        # Junction frame parser
        def _inframe():
            x = {'in-frame': 'T', 'out-of-frame': 'F'}
            return x.get(summary['JUNCTION frame'], None)

        # Stop codon parser
        def _stop():
            return 'T' if 'stop codon' in summary['Functionality comment'] else 'F'

        # Mutated invariant parser
        def _invariant():
            x = summary['Functionality comment']
            y = summary['V-REGION potential ins/del']
            return 'T' if ('missing' in x) or ('missing' in y) else 'F'

        # Mutated invariant parser
        def _indels():
            x = summary['V-REGION potential ins/del']
            y = summary['V-REGION insertions']
            z = summary['V-REGION deletions']
            return 'T' if any([x, y, z]) else 'F'

        result = {}
        # Parse functionality information
        if 'No results' not in summary['Functionality']:
            result['rev_comp'] = _revcomp()
            result['functional'] = _functional()
            result['in_frame'] = _inframe()
            result['stop'] = _stop()
            result['mutated_invariant'] = _invariant()
            result['indels'] = _indels()

        return result

    @staticmethod
    def _parseGenes(summary):
        """
        Parse gene calls

        Arguments:
          summary : dictionary containing one row of the '1_Summary' file.

        Returns:
          dict : database entries for gene calls.
        """
        clean_regex = re.compile(r'(,)|(\(see comment\))')
        delim_regex = re.compile(r'\sor\s')

        # Gene calls
        v_str = summary['V-GENE and allele']
        d_str = summary['D-GENE and allele']
        j_str = summary['J-GENE and allele']
        v_call = delim_regex.sub(',', clean_regex.sub('', v_str)) if v_str else None
        d_call = delim_regex.sub(',', clean_regex.sub('', d_str)) if d_str else None
        j_call = delim_regex.sub(',', clean_regex.sub('', j_str)) if j_str else None

        # Locus
        locus_list = [getLocus(v_call, action='first'),
                      getLocus(j_call, action='first')]
        locus = set(filter(None, locus_list))

        # Result
        result = {'v_call': v_call,
                  'd_call': d_call,
                  'j_call': j_call,
                  'locus': locus.pop() if len(locus) == 1 else None}

        return result

    @staticmethod
    def _parseSequences(gapped, ntseq):
        """
        Parses full length V(D)J sequences

        Arguments:
          gapped : dictionary containing one row of the '2_IMGT-gapped-nt-sequences' file.
          ntseq: dictionary containing one row of the '3_Nt-sequences' file.

        Returns:
          dict : database entries for fill length V(D)J sequences.
        """
        result = {}

        # Extract ungapped sequences
        if ntseq['V-D-J-REGION']:
            result['sequence_vdj'] = ntseq['V-D-J-REGION']
        elif ntseq['V-J-REGION']:
            result['sequence_vdj'] = ntseq['V-J-REGION']
        else:
            result['sequence_vdj'] = ntseq['V-REGION']

        # Extract gapped sequences
        if gapped['V-D-J-REGION']:
            result['sequence_imgt'] = gapped['V-D-J-REGION']
        elif gapped['V-J-REGION']:
            result['sequence_imgt'] = gapped['V-J-REGION']
        else:
            result['sequence_imgt'] = gapped['V-REGION']

        return result

    @staticmethod
    def _parseVPos(gapped, ntseq):
        """
        Parses V alignment positions

        Arguments:
          gapped : dictionary containing one row of the '2_IMGT-gapped-nt-sequences' file.
          ntseq: dictionary containing one row of the '3_Nt-sequences' file.

        Returns:
          dict : database entries for V query and germline alignment positions.
        """
        result = {}
        result['v_seq_start'] = ntseq['V-REGION start']
        result['v_seq_length'] = len(ntseq['V-REGION']) if ntseq['V-REGION'] else 0
        result['v_germ_start_imgt'] = 1
        result['v_germ_length_imgt'] = len(gapped['V-REGION']) if gapped['V-REGION'] else 0

        return result

    @staticmethod
    def _parseJuncPos(junction, db):
        """
        Parses junction N/P and D alignment positions

        Arguments:
          junction : dictionary containing one row of the '6_Junction' file.
          db : database containing V alignment information.

        Returns:
          dict : database entries for junction, N/P and D region alignment positions.
        """
        v_start = db['v_seq_start']
        v_length = db['v_seq_length']

        # First N/P length
        def _np1():
            nb = [junction['P3\'V-nt nb'],
                  junction['N-REGION-nt nb'],
                  junction['N1-REGION-nt nb'],
                  junction['P5\'D-nt nb']]
            return sum(int(i) for i in nb if i)

        # D start
        def _dstart():
            nb = [v_start,
                  v_length,
                  junction['P3\'V-nt nb'],
                  junction['N-REGION-nt nb'],
                  junction['N1-REGION-nt nb'],
                  junction['P5\'D-nt nb']]
            return sum(int(i) for i in nb if i)

        # Second N/P length
        def _np2():
            nb = [junction['P3\'D-nt nb'],
                  junction['N2-REGION-nt nb'],
                  junction['P5\'J-nt nb']]
            return sum(int(i) for i in nb if i)

        result = {}
        # Junction sequence
        result['junction'] = junction['JUNCTION']
        result['junction_aa'] = junction['JUNCTION (AA)']
        result['junction_length'] = len(junction['JUNCTION']) if junction['JUNCTION'] else 0

        # N/P and D alignment positions
        result['np1_length'] = _np1()
        result['d_seq_start'] = _dstart()
        result['d_seq_length'] = int(junction['D-REGION-nt nb'] or 0)
        result['d_germ_start'] = int(junction['5\'D-REGION trimmed-nt nb'] or 0) + 1
        result['d_germ_length'] = int(junction['D-REGION-nt nb'] or 0)
        result['np2_length'] = _np2()

        return result

    @staticmethod
    def _parseJPos(gapped, ntseq, junction, db):
        """
        Parses J alignment positions

        Arguments:
          gapped : dictionary containing one row of the '2_IMGT-gapped-nt-sequences' file.
          ntseq: dictionary containing one row of the '3_Nt-sequences' file.
          junction : dictionary containing one row of the '6_Junction' file.
          db : database containing V, N/P and D alignment information.

        Returns:
          dict : database entries for J region alignment positions.
        """

        # J start
        def _jstart():
            nb = [db['v_seq_start'],
                  db['v_seq_length'],
                  db['np1_length'],
                  db['d_seq_length'],
                  db['np2_length']]
            return sum(int(i) for i in nb if i)

        # J region alignment positions
        result = {}
        result['j_seq_start'] = _jstart()
        result['j_seq_length'] = len(ntseq['J-REGION']) if ntseq['J-REGION'] else 0
        result['j_germ_start'] = int(junction['5\'J-REGION trimmed-nt nb'] or 0) + 1
        result['j_germ_length'] = len(gapped['J-REGION']) if gapped['J-REGION'] else 0

        return result

    @staticmethod
    def _parseScores(summary):
        """
        Parse alignment scores

        Arguments:
          summary : dictionary containing one row of the '1_Summary' file.

        Returns:
          dict : database entries for alignment scores.
        """
        result = {}

        # V score
        try:
            result['v_score'] = float(summary['V-REGION score'])
        except (TypeError, ValueError):
            result['v_score'] = None
        # V identity
        try:
            result['v_identity'] = float(summary['V-REGION identity %']) / 100.0
        except (TypeError, ValueError):
            result['v_identity'] = 'None'
        # J score
        try:
            result['j_score'] = float(summary['J-REGION score'])
        except (TypeError, ValueError):
            result['j_score'] = None
        # J identity
        try:
            result['j_identity'] = float(summary['J-REGION identity %']) / 100.0
        except (TypeError, ValueError):
            result['j_identity'] = None

        return result

    @staticmethod
    def _parseJuncDetails(junction):
        """
        Parse detailed junction region information

        Arguments:
          junction : dictionary containing one row of the '6_Junction' file.

        Returns:
          dict : database entries for detailed D, N and P region information.
        """

        # D reading frame
        def _dframe():
            frame = None
            x = junction['D-REGION reading frame']
            if x:
                try:
                    frame = int(x)
                except ValueError:
                    m = re.search(r'reading frame ([0-9])', x).group(1)
                    frame = int(m)
            return frame

        # First N region length
        def _n1():
            nb = [junction['N-REGION-nt nb'], junction['N1-REGION-nt nb']]
            return sum(int(i) for i in nb if i)

        # D Frame and junction fields
        result = {}
        result['d_frame'] = _dframe()
        result['n1_length'] = _n1()
        result['n2_length'] = int(junction['N2-REGION-nt nb'] or 0)
        result['p3v_length'] = int(junction['P3\'V-nt nb'] or 0)
        result['p5d_length'] = int(junction['P5\'D-nt nb'] or 0)
        result['p3d_length'] = int(junction['P3\'D-nt nb'] or 0)
        result['p5j_length'] = int(junction['P5\'J-nt nb'] or 0)

        return result

    def parseRecord(self, summary, gapped, ntseq, junction):
        """
        Parses a single row from each IMTG file.

        Arguments:
          summary : dictionary containing one row of the '1_Summary' file.
          gapped : dictionary containing one row of the '2_IMGT-gapped-nt-sequences' file.
          ntseq : dictionary containing one row of the '3_Nt-sequences' file.
          junction : dictionary containing one row of the '6_Junction' file.

        Returns:
          dict: database entry for the row.
        """
        # Check that rows are syncronized
        id_set = [summary['Sequence ID'],
                  gapped['Sequence ID'],
                  ntseq['Sequence ID'],
                  junction['Sequence ID']]
        if len(set(id_set)) != 1:
            printError('IMGT files are corrupt starting with Summary file record %s.' % id_set[0])

        # Initialize db with query ID and sequence
        db = {'sequence_id': summary['Sequence ID'],
              'sequence_input': summary['Sequence']}

        # Parse required fields
        db.update(IMGTReader._parseFunctionality(summary))
        db.update(IMGTReader._parseGenes(summary))
        db.update(IMGTReader._parseSequences(gapped, ntseq))
        db.update(IMGTReader._parseVPos(gapped, ntseq))
        db.update(IMGTReader._parseJuncPos(junction, db))
        db.update(IMGTReader._parseJPos(gapped, ntseq, junction, db))

        # Parse optional fields
        db.update(IMGTReader._parseScores(summary))
        rd = RegionDefinition(junction_length=db.get('junction_length', None),
                              amino_acid=False, definition='default')
        db.update(rd.getRegions(db.get('sequence_imgt', None)))
        db.update(IMGTReader._parseJuncDetails(junction))

        return db

    def __iter__(self):
        """
        Iterator initializer.

        Returns:
          changeo.IO.IMGTReader
        """
        return self

    def __next__(self):
        """
        Next method.

        Returns:
          changeo.Receptor.Receptor : parsed IMGT/HighV-QUEST result as an Receptor (receptor=True) or dictionary (receptor=False).
        """
        # Get next set of records from dictionary readers
        try:
            summary, gapped, ntseq, junction = next(self.records)
        except StopIteration:
            raise StopIteration

        db = self.parseRecord(summary, gapped, ntseq, junction)

        if self.receptor:
            return Receptor(db)
        else:
            return db


class IgBLASTReader:
    """
    An iterator to read and parse IgBLAST output files
    """
    # Ordered list of known fields
    @staticmethod
    def customFields(schema=None):
        """
        Returns non-standard fields defined by the parser

        Arguments:
          schema : schema class to pass field through for conversion.
                   If None, return changeo.Receptor.Receptor attribute names.

        Returns:
          list : list of field names.
        """
        # IgBLAST scoring fields
        fields = ['v_score',
                  'v_identity',
                  'v_evalue',
                  'v_cigar',
                  'd_score',
                  'd_identity',
                  'd_evalue',
                  'd_cigar',
                  'j_score',
                  'j_identity',
                  'j_evalue',
                  'j_cigar',
                  'fwr1_imgt',
                  'fwr2_imgt',
                  'fwr3_imgt',
                  'fwr4_imgt',
                  'cdr1_imgt',
                  'cdr2_imgt',
                  'cdr3_imgt']

        # Convert field names if schema provided
        if schema is not None:
            fields = [schema.fromReceptor(f) for f in fields]

        return fields

    def __init__(self, igblast, sequences, references, asis_calls=False, regions='default', receptor=True,
                 infer_junction=False):
        """
        Initializer.

        Arguments:
          igblast (file): handle to an open IgBLAST output file written with '-outfmt 7 std qseq sseq btop'.
          sequences (dict): dictionary of query sequences;
                            sequence descriptions as keys with original query sequences as SeqRecord values.
          references (dict): dictionary of IMGT gapped germline sequences.
          asis_calls (bool): if True do not parse gene calls for allele names.
          regions (str): name of the IMGT FWR/CDR region definitions to use.
          receptor (bool): if True (default) iteration returns an Receptor object, otherwise it returns a dictionary.
          infer_junction (bool): if True, infer the junction region if not reported by IgBLAST.

        Returns:
          changeo.IO.IgBLASTReader
        """
        # Arguments
        self.igblast = igblast
        self.sequences = sequences
        self.references = references
        self.regions = regions
        self.asis_calls = asis_calls
        self.receptor = receptor
        self.infer_junction = infer_junction

        # Define parsing blocks
        self.groups = groupby(self.igblast, lambda x: not re.match('# IGBLAST', x))

    def _parseQueryChunk(self, chunk):
        """
        Parse query section

        Arguments:
          chunk : list of strings

        Returns:
          str : query identifier
        """
        # Extract query id from comments
        query = next((x for x in chunk if x.startswith('# Query:')))

        return query.replace('# Query: ', '', 1)

    def _parseSummaryChunk(self, chunk):
        """
        Parse summary section

        Args:
            chunk: list of strings

        Returns:
            dict : summary section.
        """
        # Mapping for field names in the summary section
        summary_map = {'Top V gene match': 'v_match',
                       'Top D gene match': 'd_match',
                       'Top J gene match': 'j_match',
                       'Top C gene match': 'c_match',
                       'Chain type': 'chain_type',
                       'stop codon': 'stop_codon',
                       'V-J frame': 'vj_frame',
                       'Productive': 'productive',
                       'Strand': 'strand',
                       'V Frame shift': 'v_frameshift'}

        # Extract column names from comments
        f = next((x for x in chunk if x.startswith('# V-(D)-J rearrangement summary')))
        f = re.search(r'summary for query sequence \((.+)\)\.', f).group(1)
        columns = [summary_map[x.strip()] for x in f.split(',')]

        # Extract first row as a list
        row = next((x.split('\t') for x in chunk if not x.startswith('#')))

        # Populate template dictionary with parsed fields
        summary = {v: None for v in summary_map.values()}
        summary.update(dict(zip(columns, row)))

        return summary

    def _parseSubregionChunk(self, chunk):
        """
        Parse CDR3 sequences generated by IgBLAST

        Args:
          chunk: list of strings

        Returns:
          dict : nucleotide and amino acid CDR3 sequences
        """
        # Example:
        #   # Sub-region sequence details (nucleotide sequence, translation, start, end)
        #   CDR3  CAACAGTGGAGTAGTTACCCACGGACG QQWSSYPRT	248	287

        # Define column names
        cdr3_map = {'nucleotide sequence': 'cdr3_igblast',
                    'translation': 'cdr3_igblast_aa',
                    'start': 'cdr3_igblast_start',
                    'end': 'cdr3_igblast_end'}

        # Extract column names from comments
        f = next((x for x in chunk if x.startswith('# Sub-region sequence details')))
        f = re.search(r'sequence details \((.+)\)', f).group(1)
        columns = [cdr3_map[x.strip()] for x in f.split(',')]

        # Extract first CDR3 as a list and remove the CDR3 label
        rows = next((x.split('\t') for x in chunk if x.startswith('CDR3')))[1:]

        # Populate dictionary with parsed fields
        cdr = {v: None for v in columns}
        cdr.update(dict(zip(columns, rows)))

        # Add length
        if cdr.get('cdr3_igblast', None) is not None:
            cdr['cdr3_igblast_length'] = len(cdr['cdr3_igblast'])

        return cdr

    def _parseHitsChunk(self, chunk):
        """
        Parse hits section

        Args:
          chunk: list of strings

        Returns:
          list: hit table as a list of dictionaries
        """
        # Extract column names from comments
        f = next((x for x in chunk if x.startswith('# Fields:')))
        columns = chain(['segment'], f.replace('# Fields:', '', 1).split(','))
        columns = [x.strip() for x in columns]
        # Split non-comment rows into a list of lists
        rows = [x.split('\t') for x in chunk if not x.startswith('#')]
        # Create list of dictionaries containing hits
        hits = [{k: x[i] for i, k in enumerate(columns)} for x in rows]

        return hits

    def _parseSummarySection(self, summary, db, asis_calls=False):
        """
        Parse summary section

        Arguments:
          summary :  summary section dictionary return by parseBlock
          db : initial database dictionary.
          asis_calls : if True do not parse gene calls for allele names.

        Returns:
          dict : db of results.
        """
        result = {}
        # Parse V, D, and J calls
        if not asis_calls:
            v_call = getVAllele(summary['v_match'], action='list')
            d_call = getDAllele(summary['d_match'], action='list')
            j_call = getJAllele(summary['j_match'], action='list')
            c_call = getCAllele(summary['c_match'], action='list')
            result['v_call'] = ','.join(v_call) if v_call else None
            result['d_call'] = ','.join(d_call) if d_call else None
            result['j_call'] = ','.join(j_call) if j_call else None
            result['c_call'] = ','.join(c_call) if c_call else None
        else:
            result['v_call'] = None if summary['v_match'] == 'N/A' else summary['v_match']
            result['d_call'] = None if summary['d_match'] == 'N/A' else summary['d_match']
            result['j_call'] = None if summary['j_match'] == 'N/A' else summary['j_match']
            result['c_call'] = None if summary['c_match'] == 'N/A' else summary['c_match']

        # Parse locus
        locus = None if summary['chain_type'] == 'N/A' else summary['chain_type']
        locus_map = {'VH': 'IGH', 'VK': 'IGK', 'VL': 'IGL',
                     'VB': 'TRB', 'VD': 'TRD', 'VA': 'TRA', 'VG': 'TRG'}
        result['locus'] = locus_map.get(locus, locus)

        # Parse quality information
        result['stop'] = 'T' if summary['stop_codon'] == 'Yes' else 'F'
        result['in_frame'] = 'T' if summary['vj_frame'] == 'In-frame' else 'F'
        result['functional'] = 'T' if summary['productive'] == 'Yes' else 'F'

        # Reverse complement input sequence if required
        if summary['strand'] == '-':
            seq_rc = Seq(db['sequence_input']).reverse_complement()
            result['sequence_input'] = str(seq_rc)
            result['rev_comp'] = 'T'
        else:
            result['rev_comp'] = 'F'

        # Add v_frameshift field if present
        if 'v_frameshift' in summary:
            result['v_frameshift'] = 'T' if summary['v_frameshift'] == 'Yes' else 'F'

        return result

    def _parseSubregionSection(self, section, sequence):
        """
        Parse subregion section

        Arguments:
          section :  subregion section dictionary return by parseBlock
          sequence : input sequence

        Returns:
          dict : db of results.
        """
        # Extract junction
        junc_start = int(section['cdr3_igblast_start']) - 3
        junc_end = int(section['cdr3_igblast_end']) + 3
        junc_seq = sequence[(junc_start - 1):junc_end]
        junc_len = len(junc_seq)

        # Translation
        junc_tmp = junc_seq.replace('-', 'N').replace('.', 'N')
        if junc_len % 3 > 0:  junc_tmp = junc_tmp[:junc_len - junc_len % 3]
        junc_aa = str(Seq(junc_tmp).translate())

        # Build return values
        return {'junction': junc_seq,
                'junction_aa': junc_aa,
                'junction_length': junc_len,
                'junction_start': junc_start,
                'junction_end': junc_end}

    def _parseVHitPos(self, v_hit):
        """
        Parse V alignment positions

        Arguments:
          v_hit :  V alignment row from the hit table

        Returns:
          dict: db of D starts and lengths
        """
        result = {}
        # Germline positions
        result['v_germ_start_vdj'] = int(v_hit['s. start'])
        result['v_germ_length_vdj'] = int(v_hit['s. end']) - result['v_germ_start_vdj'] + 1
        # Query sequence positions
        result['v_seq_start'] = int(v_hit['q. start'])
        result['v_seq_length'] = int(v_hit['q. end']) - result['v_seq_start'] + 1
        result['indels'] = 'F' if int(v_hit['gap opens']) == 0 else 'T'

        return result

    def _parseDHitPos(self, d_hit, overlap):
        """
        Parse D alignment positions

        Arguments:
          d_hit :  D alignment row from the hit table
          overlap : V-D overlap length

        Returns:
          dict: db of D starts and lengths
        """
        result = {}
        # Query sequence positions
        result['d_seq_start'] = int(d_hit['q. start']) + overlap
        result['d_seq_length'] = max(int(d_hit['q. end']) - result['d_seq_start'] + 1, 0)
        # Germline positions
        result['d_germ_start'] = int(d_hit['s. start']) + overlap
        result['d_germ_length'] = max(int(d_hit['s. end']) - result['d_germ_start'] + 1, 0)

        return result

    def _parseJHitPos(self, j_hit, overlap):
        """
        Parse J alignment positions

        Arguments:
          j_hit :  J alignment row from the hit table
          overlap : D-J or V-J overlap length

        Returns:
          dict: db of J starts and lengths
        """
        result = {}
        result['j_seq_start'] = int(j_hit['q. start']) + overlap
        result['j_seq_length'] = max(int(j_hit['q. end']) - result['j_seq_start'] + 1, 0)
        result['j_germ_start'] = int(j_hit['s. start']) + overlap
        result['j_germ_length'] = max(int(j_hit['s. end']) - result['j_germ_start'] + 1, 0)

        return result

    def _appendSeq(self, seq, hits, start, trim=True):
        """
        Append aligned query sequence segment

        Arguments:
          seq :  sequence to modify.
          hits : hit table row for the sequence.
          start : start position of the query sequence.
          trim : if True then remove insertions from the hit sequence before appending.

        Returns:
          str: modified sequence.
        """
        if 'subject seq' not in hits or 'query seq' not in hits:
            return None

        # Remove insertions
        if trim:
            for m in re.finditer(r'-', hits['subject seq']):
                ins = m.start()
                seq += hits['query seq'][start:ins]
                start = ins + 1

        # Append
        seq += hits['query seq'][start:]

        return seq

    def _parseVHits(self, hits, db):
        """
        Parse V hit sub-table

        Arguments:
          hits :  hit table as a list of dictionaries.
          db : database dictionary containing summary results.

        Returns:
          dict : db of results.
        """
        result = {}
        seq_vdj = db['sequence_vdj']
        seq_trim = db['sequence_trim']
        v_hit = next(x for x in hits if x['segment'] == 'V')

        # Alignment positions
        result.update(self._parseVHitPos(v_hit))
        # Update VDJ sequence with and without removing insertions
        result['sequence_vdj'] = self._appendSeq(seq_vdj, v_hit, 0, trim=False)
        result['sequence_trim'] = self._appendSeq(seq_trim, v_hit, 0, trim=True)

        return result

    def _parseDHits(self, hits, db):
        """
        Parse D hit sub-table

        Arguments:
          hits :  hit table as a list of dictionaries.
          db : database dictionary containing summary and V results.

        Returns:
          dict : db of results.
        """
        result = {}
        seq_vdj = db['sequence_vdj']
        seq_trim = db['sequence_trim']
        d_hit = next(x for x in hits if x['segment'] == 'D')

        # Determine N-region length and amount of J overlap with V or D alignment
        overlap = 0
        if db['v_call']:
            np1_len = int(d_hit['q. start']) - (db['v_seq_start'] + db['v_seq_length'])
            if np1_len < 0:
                result['np1_length'] = 0
                overlap = abs(np1_len)
            else:
                result['np1_length'] = np1_len
                np1_start = db['v_seq_start'] + db['v_seq_length'] - 1
                np1_end = int(d_hit['q. start']) - 1
                if seq_vdj is not None:
                    seq_vdj += db['sequence_input'][np1_start:np1_end]
                if seq_trim is not None:
                    seq_trim += db['sequence_input'][np1_start:np1_end]

        # D alignment positions
        result.update(self._parseDHitPos(d_hit, overlap))
        # Update VDJ sequence with and without removing insertions
        result['sequence_vdj'] = self._appendSeq(seq_vdj, d_hit, overlap, trim=False)
        result['sequence_trim'] = self._appendSeq(seq_trim, d_hit, overlap, trim=True)

        return result


    def _parseJHits(self, hits, db):
        """
        Parse J hit sub-table

        Arguments:
          hits :  hit table as a list of dictionaries.
          db : database dictionary containing summary, V and D results.

        Returns:
          dict : db of results.
        """
        result = {}
        seq_vdj = db['sequence_vdj']
        seq_trim = db['sequence_trim']
        j_hit = next(x for x in hits if x['segment'] == 'J')

        # Determine N-region length and amount of J overlap with V or D alignment
        overlap = 0
        if db['d_call']:
            np2_len = int(j_hit['q. start']) - (db['d_seq_start'] + db['d_seq_length'])
            if np2_len < 0:
                result['np2_length'] = 0
                overlap = abs(np2_len)
            else:
                result['np2_length'] = np2_len
                n2_start = db['d_seq_start'] + db['d_seq_length'] - 1
                n2_end = int(j_hit['q. start']) - 1
                if seq_vdj is not None:
                    seq_vdj += db['sequence_input'][n2_start:n2_end]
                if seq_trim is not None:
                    seq_trim += db['sequence_input'][n2_start:n2_end]
        elif db['v_call']:
            np1_len = int(j_hit['q. start']) - (db['v_seq_start'] + db['v_seq_length'])
            if np1_len < 0:
                result['np1_length'] = 0
                overlap = abs(np1_len)
            else:
                result['np1_length'] = np1_len
                np1_start = db['v_seq_start'] + db['v_seq_length'] - 1
                np1_end = int(j_hit['q. start']) - 1
                if seq_vdj is not None:
                    seq_vdj += db['sequence_input'][np1_start: np1_end]
                if seq_trim is not None:
                    seq_trim += db['sequence_input'][np1_start: np1_end]
        else:
            result['np1_length'] = 0

        # J alignment positions
        result.update(self._parseJHitPos(j_hit, overlap))
        # Update VDJ sequence with and without removing insertions
        result['sequence_vdj'] = self._appendSeq(seq_vdj, j_hit, overlap, trim=False)
        result['sequence_trim'] = self._appendSeq(seq_trim, j_hit, overlap, trim=True)

        return result

    def _parseHitScores(self, hits, segment):
        """
        Parse alignment scores

        Arguments:
          hits :  hit table as a list of dictionaries.
          segment : segment name; one of 'v', 'd' or 'j'.

        Returns:
          dict : scores
        """
        result = {}
        s_hit = next(x for x in hits if x['segment'] == segment.upper())

        # Score
        try:
            result['%s_score' % segment] = float(s_hit['bit score'])
        except (TypeError, ValueError):
            result['%s_score' % segment] = None
        # Identity
        try:
            result['%s_identity' % segment] = float(s_hit['% identity']) / 100.0
        except (TypeError, ValueError):
            result['%s_identity' % segment] = None
        # E-value
        try:
            result['%s_evalue' % segment] = float(s_hit['evalue'])
        except (TypeError, ValueError):
            result['%s_evalue' % segment] = None
        # BTOP
        try:
            result['%s_btop' % segment] = s_hit['BTOP']
        except (KeyError, TypeError, ValueError):
            result['%s_btop' % segment] = None
        # CIGAR
        try:
            align = decodeBTOP(s_hit['BTOP'])
            align = padAlignment(align, int(s_hit['q. start']) - 1, int(s_hit['s. start']) - 1)
            result['%s_cigar' % segment] = encodeCIGAR(align)
        except (KeyError, TypeError, ValueError):
            result['%s_cigar' % segment] = None

        return result

    def parseBlock(self, block):
        """
        Parses an IgBLAST result into separate sections

        Arguments:
          block (iter): an iterator from itertools.groupby containing a single IgBLAST result.

        Returns:
          dict: a parsed results block;
                with the keys 'query' (sequence identifier as a string),
                'summary' (dictionary of the alignment summary),
                'subregion' (dictionary of IgBLAST CDR3 sequences), and
                'hits' (VDJ hit table as a list of dictionaries).
                Returns None if the block has no data that can be parsed.
        """
        # Parsing info
        #
        #   Columns for non-hit-table sections
        #     'V-(D)-J rearrangement summary': (Top V gene match, Top D gene match, Top J gene match, Chain type, stop codon, V-J frame, Productive, Strand)
        #     'V-(D)-J junction details': (V end, V-D junction, D region, D-J junction, J start)
        #     'Alignment summary': (from, to, length, matches, mismatches, gaps, percent identity)
        #     'subregion': (nucleotide sequence, translation, start, end)
        #
        #   Ignored sections
        #     'junction': '# V-(D)-J junction details'
        #     'v_alignment': '# Alignment summary'
        #
        #   Hit table fields for -outfmt "7 std qseq sseq btop"
        #     0:  segment
        #     1:  query id
        #     2:  subject id
        #     3:  % identity
        #     4:  alignment length
        #     5:  mismatches
        #     6:  gap opens
        #     7:  gaps
        #     8:  q. start
        #     9:  q. end
        #    10:  s. start
        #    11:  s. end
        #    12:  evalue
        #    13:  bit score
        #    14:  query seq
        #    15:  subject seq
        #    16:  btop
        # Map of valid block parsing keys and functions
        chunk_map = {'query': ('# Query:', self._parseQueryChunk),
                     'summary': ('# V-(D)-J rearrangement summary', self._parseSummaryChunk),
                     'subregion': ('# Sub-region sequence details', self._parseSubregionChunk),
                     'hits': ('# Hit table', self._parseHitsChunk)}

        # Parsing chunks
        results = {}
        for match, chunk in groupby(block, lambda x: x != '\n'):
            if match:
                # Strip whitespace and convert to list
                chunk = [x.strip() for x in chunk]

                # Parse non-query sections
                chunk_dict = {k: f(chunk) for k, (v, f) in chunk_map.items() if chunk[0].startswith(v)}
                results.update(chunk_dict)

        return results if results else None

    def parseSections(self, sections):
        """
        Parses an IgBLAST sections into a db dictionary

        Arguments:
            sections : dictionary of parsed sections from parseBlock.

        Returns:
          dict : db entries.
        """
        # Initialize dictionary with input sequence and id
        db = {}
        if 'query' in sections:
            query = sections['query']
            db['sequence_id'] = query
            db['sequence_input'] = str(self.sequences[query].seq)

        # Parse summary section
        if 'summary' in sections:
            db.update(self._parseSummarySection(sections['summary'], db, asis_calls=self.asis_calls))

        # Parse hit table
        if 'hits' in sections:
            db['sequence_vdj'] = ''
            db['sequence_trim'] = ''
            if db['v_call']:
                db.update(self._parseVHits(sections['hits'], db))
                db.update(self._parseHitScores(sections['hits'], 'v'))
            if db['d_call']:
                db.update(self._parseDHits(sections['hits'], db))
                db.update(self._parseHitScores(sections['hits'], 'd'))
            if db['j_call']:
                db.update(self._parseJHits(sections['hits'], db))
                db.update(self._parseHitScores(sections['hits'], 'j'))

        # Create IMGT-gapped sequence
        if ('v_call' in db and db['v_call']) and ('sequence_trim' in db and db['sequence_trim']):
            try:
                imgt_dict = gapV(db['sequence_trim'],
                                 v_germ_start=db['v_germ_start_vdj'],
                                 v_germ_length=db['v_germ_length_vdj'],
                                 v_call=db['v_call'],
                                 references=self.references,
                                 asis_calls=self.asis_calls)
            except KeyError as e:
                imgt_dict = {'sequence_imgt': None,
                             'v_germ_start_imgt': None,
                             'v_germ_length_imgt': None}
                printWarning(e)
            db.update(imgt_dict)
            del db['sequence_trim']

        # Add junction
        if 'subregion' in sections and 'cdr3_igblast_start' in sections['subregion']:
            junc_dict = self._parseSubregionSection(sections['subregion'], db['sequence_input'])
            db.update(junc_dict)
        elif self.infer_junction and ('j_call' in db and db['j_call']) and ('sequence_imgt' in db and db['sequence_imgt']):
            junc_dict = inferJunction(db['sequence_imgt'],
                                      j_germ_start=db['j_germ_start'],
                                      j_germ_length=db['j_germ_length'],
                                      j_call=db['j_call'],
                                      references=self.references,
                                      asis_calls=self.asis_calls,
                                      regions=self.regions)
            db.update(junc_dict)

        # Add IgBLAST CDR3 sequences
        if 'subregion' in sections:
            # Sequences already parsed into dict by parseBlock
            db.update(sections['subregion'])
        else:
            # Section does not exist (ie, older version of IgBLAST or CDR3 not found)
            db.update({'cdr3_igblast': None, 'cdr3_igblast_aa': None})

        # Add FWR and CDR regions
        rd = RegionDefinition(junction_length=db.get('junction_length', None),
                              amino_acid=False, definition=self.regions)
        db.update(rd.getRegions(db.get('sequence_imgt', None)))

        return db

    def __iter__(self):
        """
        Iterator initializer.

        Returns:
          changeo.IO.IgBLASTReader
        """
        return self

    def __next__(self):
        """
        Next method.

        Returns:
          changeo.Receptor.Receptor : parsed IMGT/HighV-QUEST result as an Receptor (receptor=True) or dictionary (receptor=False).
        """
        # Get next block from groups iterator
        try:
            match = False
            block = None
            while not match:
                match, block = next(self.groups)
        except StopIteration:
            raise StopIteration

        # Parse block
        sections = self.parseBlock(block)
        db = self.parseSections(sections)

        if self.receptor:
            return Receptor(db)
        else:
            return db


class IgBLASTReaderAA(IgBLASTReader):
    """
    An iterator to read and parse IgBLAST amino acid alignment output files
    """
    @staticmethod
    def customFields(schema=None):
        """
        Returns non-standard fields defined by the parser

        Arguments:
          schema : schema class to pass field through for conversion.
                   If None, return changeo.Receptor.Receptor attribute names.

        Returns:
          list : list of field names.
        """
        # IgBLAST scoring fields
        fields = ['v_score',
                  'v_identity',
                  'v_evalue',
                  'v_cigar',
                  'fwr1_aa_imgt',
                  'fwr2_aa_imgt',
                  'fwr3_aa_imgt',
                  'cdr1_aa_imgt',
                  'cdr2_aa_imgt']

        # Convert field names if schema provided
        if schema is not None:
            fields = [schema.fromReceptor(f) for f in fields]

        return fields

    def _parseVHitPos(self, v_hit):
        """
        Parse V alignment positions

        Arguments:
          v_hit :  V alignment row from the hit table

        Returns:
          dict: db of D starts and lengths
        """
        result = {}
        # Germline positions
        result['v_germ_aa_start_vdj'] = int(v_hit['s. start'])
        result['v_germ_aa_length_vdj'] = int(v_hit['s. end']) - result['v_germ_aa_start_vdj'] + 1
        # Query sequence positions
        result['v_seq_aa_start'] = int(v_hit['q. start'])
        result['v_seq_aa_length'] = int(v_hit['q. end']) - result['v_seq_aa_start'] + 1
        result['indels'] = 'F' if int(v_hit['gap opens']) == 0 else 'T'

        return result

    def _parseVHits(self, hits, db):
        """
        Parse V hit sub-table

        Arguments:
          hits :  hit table as a list of dictionaries.
          db : database dictionary containing summary results.

        Returns:
          dict : db of results.
        """
        result = {}
        seq_vdj = db['sequence_aa_vdj']
        seq_trim = db['sequence_aa_trim']
        v_hit = next(x for x in hits if x['segment'] == 'V')

        # Alignment positions
        result.update(self._parseVHitPos(v_hit))

        # Assign V gene and update VDJ sequence with and without removing insertions
        result['v_call'] = v_hit['subject id']
        result['sequence_aa_vdj'] = self._appendSeq(seq_vdj, v_hit, 0, trim=False)
        result['sequence_aa_trim'] = self._appendSeq(seq_trim, v_hit, 0, trim=True)

        # Derived functionality
        result['locus'] = getLocus(result['v_call'], action='first')
        result['stop'] = '*' in result['sequence_aa_vdj']

        return result

    def parseSections(self, sections):
        """
        Parses an IgBLAST sections into a db dictionary

        Arguments:
            sections : dictionary of parsed sections from parseBlock.

        Returns:
          dict : db entries.
        """
        # Initialize dictionary with input sequence and id
        db = {}
        if 'query' in sections:
            query = sections['query']
            db['sequence_id'] = query
            db['sequence_aa_input'] = str(self.sequences[query].seq)

        # Parse hit table
        if 'hits' in sections:
            db['v_call'] = ''
            db['sequence_aa_vdj'] = ''
            db['sequence_aa_trim'] = ''
            db.update(self._parseVHits(sections['hits'], db))
            db.update(self._parseHitScores(sections['hits'], 'v'))

        # Create IMGT-gapped sequence
        if ('v_call' in db and db['v_call']) and ('sequence_aa_trim' in db and db['sequence_aa_trim']):
            try:
                gap = gapV(db['sequence_aa_trim'],
                           v_germ_start=db['v_germ_aa_start_vdj'],
                           v_germ_length=db['v_germ_aa_length_vdj'],
                           v_call=db['v_call'],
                           references=self.references,
                           asis_calls=self.asis_calls)
                imgt_dict = {'sequence_aa_imgt': gap['sequence_imgt'],
                             'v_germ_aa_start_imgt': gap['v_germ_start_imgt'],
                             'v_germ_aa_length_imgt': gap['v_germ_length_imgt']}
            except KeyError as e:
                imgt_dict = {'sequence_aa_imgt': None,
                             'v_germ_aa_start_imgt': None,
                             'v_germ_aa_length_imgt': None}
                printWarning(e)
            db.update(imgt_dict)
            del db['sequence_aa_trim']

        # Add FWR and CDR regions
        rd = RegionDefinition(junction_length=db.get('junction_length', None),
                              amino_acid=True, definition=self.regions)
        regions = rd.getRegions(db.get('sequence_aa_imgt', None))
        regions = {'fwr1_aa_imgt': regions['fwr1_imgt'],
                   'fwr2_aa_imgt': regions['fwr2_imgt'],
                   'fwr3_aa_imgt': regions['fwr3_imgt'],
                   'cdr1_aa_imgt': regions['cdr1_imgt'],
                   'cdr2_aa_imgt': regions['cdr2_imgt']}
        db.update(regions)

        return db


class IHMMuneReader:
    """
    An iterator to read and parse iHMMune-Align output files.
    """
    # iHMMuneAlign columns
    # Courtesy of Katherine Jackson
    #
    #  1: Identifier - sequence identifer from FASTA input file
    #  2: IGHV - IGHV gene match from the IGHV repertoire, if multiple genes had equally
    #            good alignments both will be listed, if indels were found this will be
    #            listed, in case of multiple IGHV all further data is reported with
    #            respect to the first listed gene
    #  3: IGHD - IGHD gene match, if no IGHD could be found or the IGHD that was found
    #            failed to meet confidence criteria this will be 'NO_DGENE_ALIGNMENT'
    #  4: IGHJ - IGHJ gene match, only a single best matching IGHJ is reported, if indels
    #            are found then 'indel' will be listed
    #  5: V-REGION - portion of input sequence that matches to the germline IGHV, were
    #                nucleotide are missing at start or end the sequence is padded back
    #                to full length with '.' (the exonuclease loss from the end of the
    #                gene will therefore be equal to the number of '.' characters at the
    #                5` end), mismatches between germline and rearranged are in uppercase,
    #                matches are in lowercase
    #  6: N1-REGION - sequence between V- and D-REGIONs
    #  7: D-REGION - portion of input sequence that matches to the germline IGHD
    #                (model doesn't currently permit indels in the IGHD), where IGHD is
    #                reported as 'NO_DGENE_ALIGNMENT' this field contains all nucleotides
    #                between the V- and J-REGIONs
    #  8: N2-REGION - sequence between D- and J-REGIONs
    #  9: J-REGION - portion of the input sequence that matches germline IGHJ, padded
    #                5` and 3` to length of germline match
    # 10: V mutation count - count of mismatches in the V-REGION
    # 11: D mutation count - count of mismatches in the D-REGION
    # 12: J mutation count - count of mismatches in the J-REGION
    # 13: count of ambigious nts - count of 'n' or 'x' nucleotides in the input sequence
    # 14: IGHJ in-frame - 'true' is IGHJ is in-frame and 'false' if IGHJ is out-of-frame,
    #                     WARNING indels and germline IGHV database sequences that are
    #                     not RF1 can cause this to report inaccurately
    # 15: IGHV start offset - offset for start of alignment between input sequence and
    #                         germline IGHV
    #                         NOTE: appears to be base 1 indexing.
    # 16: stop codons - count of stop codons in the sequence, WARNING indels and germline
    #                   IGHV database sequence that are not RF can cause this to be inaccurate
    # 17: IGHD probability - probability that N-nucleotide addition could have created the
    #                        D-REGION sequence
    # 18: HMM path score - path score from HMM
    # 19: reverse complement - 0 for no reverse complement, 1 if alignment was to reverse
    #                          complement NOTE currently this version only functions with
    #                          input in coding orientation
    # 20: mutations in common region - count of mutations in common region, which is a
    #                                  portion of the IGHV that is highly conserved,
    #                                  mutations in this region are used to set various
    #                                  probabilities in the HMM
    # 21: ambigious nts in common region - count of 'n' or 'x' nucleotides in the
    #                                      common region
    # 22: IGHV start offset  - offset for start of alignment between input sequence and
    #                          germline IGHV
    #                          NOTE: appears to be base 0 indexing.
    #                          NOTE: don't know if this differs from 15; it doesn't appear to.
    # 23: IGHV gene length - length of IGHV gene
    # 24: A score - A score probability is calculated from the common region mutations
    #               and is used for HMM calculations relating to expected mutation
    #               probability at different positions in the rearrangement
    ihmmune_fields = ['SEQUENCE_ID',
                      'V_CALL',
                      'D_CALL',
                      'J_CALL',
                      'V_SEQ',
                      'NP1_SEQ',
                      'D_SEQ',
                      'NP2_SEQ',
                      'J_SEQ',
                      'V_MUT',
                      'D_MUT',
                      'J_MUT',
                      'NX_COUNT',
                      'J_INFRAME',
                      'V_SEQ_START',
                      'STOP_COUNT',
                      'D_PROB',
                      'HMM_SCORE',
                      'RC',
                      'COMMON_MUT',
                      'COMMON_NX_COUNT',
                      'V_SEQ_START',
                      'V_SEQ_LENGTH',
                      'A_SCORE']

    # Ordered list of known fields
    @staticmethod
    def customFields(scores=False, regions=False, cell=False, schema=None):
        """
        Returns non-standard Receptor attributes defined by the parser

        Arguments:
          scores : if True include alignment scoring fields.
          regions : if True include IMGT-gapped CDR and FWR region fields.
          schema : schema class to pass field through for conversion.
                   If None, return changeo.Receptor.Receptor attribute names.

        Returns:
          list : list of field names.
        """
        # Alignment scoring fields
        score_fields = ['vdj_score']

        # FWR amd CDR fields
        region_fields = ['fwr1_imgt',
                         'fwr2_imgt',
                         'fwr3_imgt',
                         'fwr4_imgt',
                         'cdr1_imgt',
                         'cdr2_imgt',
                         'cdr3_imgt']

        fields = []
        if scores:  fields.extend(score_fields)
        if regions:  fields.extend(region_fields)

        # Convert field names if schema provided
        if schema is not None:
            fields = [schema.fromReceptor(f) for f in fields]

        return fields

    def __init__(self, ihmmune, sequences, references, receptor=True):
        """
        Initializer

        Arguments:
          ihmmune (file): handle to an open iHMMune-Align output file.
          sequences (dict): dictionary with sequence descriptions as keys mapping to the SeqRecord containing
                            the original query sequences.
          references (dict): dictionary of IMGT gapped germline sequences.
          receptor (bool): if True (default) iteration returns an Receptor object, otherwise it returns a dictionary

        Returns:
          changeo.IO.IHMMuneReader
        """
        # Arguments
        self.ihmmune = ihmmune
        self.sequences = sequences
        self.references = references
        self.receptor = receptor

        # Open reader
        self.records = csv.DictReader(self.ihmmune, fieldnames=IHMMuneReader.ihmmune_fields,
                                      delimiter=';', quotechar='"')

    @staticmethod
    def _parseFunctionality(record):
        """
        Parse functionality information

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.

        Returns:
          dict : database entries containing functionality information.
        """
        # Orientation
        def _revcomp():
            return 'F' if int(record['RC']) == 0 else 'T'

        # Functional
        def _functional():
            if not record['V_CALL'] or \
                    record['V_CALL'].startswith('NA - ') or \
                    record['J_INFRAME'] != 'true' or \
                    not record['J_CALL'] or \
                    record['J_CALL'] == 'NO_JGENE_ALIGNMENT' or \
                    int(record['STOP_COUNT']) > 0:
                return 'F'
            else:
                return 'T'

        # Stop codon
        def _stop():
            return 'T' if int(record['STOP_COUNT']) > 0 else 'F'

        # J in-frame
        def _inframe():
            return 'T' if record['J_INFRAME'] == 'true' else 'F'

        # Indels
        def _indels():
            check = [x is not None and 'indels' in x \
                     for x in [record['V_CALL'], record['D_CALL'], record['J_CALL']]]
            return 'T' if any(check) else 'F'

        # Parse functionality
        result = {'rev_comp': _revcomp(),
                  'functional': _functional(),
                  'in_frame': _inframe(),
                  'stop': _stop(),
                  'indels':  _indels()}

        return result

    @staticmethod
    def _parseGenes(record):
        """
        Parse gene calls

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.

        Returns:
          dict : database entries for gene calls.
        """
        # Extract allele calls
        v_call = getVAllele(record['V_CALL'], action='list')
        d_call = getDAllele(record['D_CALL'], action='list')
        j_call = getJAllele(record['J_CALL'], action='list')

        # Locus
        locus_list = [getLocus(record['V_CALL'], action='first'),
                      getLocus(record['J_CALL'], action='first')]
        locus = set(filter(None, locus_list))

        # Build return object
        result = {'v_call': ','.join(v_call) if v_call else None,
                  'd_call': ','.join(d_call) if d_call else None,
                  'j_call': ','.join(j_call) if j_call else None,
                  'locus': locus.pop() if len(locus) == 1 else None}

        return result

    @staticmethod
    def _parseNPHit(record):
        """
        Parse N/P region alignment information

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.

        Returns:
          dict : database entries containing N/P region lengths.
        """
        # N/P lengths
        result = {'np1_length': len(record['NP1_SEQ']),
                  'np2_length': len(record['NP2_SEQ'])}

        return result

    @staticmethod
    def _parseVHit(record, db):
        """
        Parse V alignment information

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.
          db : database containing V and D alignment information.

        Returns:
          dict : database entries containing V call and alignment positions.
        """
        # Default return
        result = {'v_seq_start': None,
                  'v_seq_length': None,
                  'v_germ_start_vdj': None,
                  'v_germ_length_vdj': None}

        # Find V positions
        if db['v_call']:
            # Query positions
            result['v_seq_start'] = int(record['V_SEQ_START'])
            result['v_seq_length'] = len(record['V_SEQ'].strip('.'))
            # Germline positions
            result['v_germ_start_vdj'] = 1
            result['v_germ_length_vdj'] = result['v_seq_length']

        return result

    def _parseDHit(record, db):
        """
        Parse D alignment information

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.
          db : database containing V alignment information.


        Returns:
          dict : database entries containing D call and alignment positions.
        """

        # D start position
        def _dstart():
            nb = [db['v_seq_start'],
                  db['v_seq_length'],
                  db['np1_length']]
            return sum(int(i) for i in nb if i)

        # Default return
        result = {'d_seq_start': None,
                  'd_seq_length': None,
                  'd_germ_start': None,
                  'd_germ_length': None}

        if db['d_call']:
            # Query positions
            result['d_seq_start'] = _dstart()
            result['d_seq_length'] = len(record['D_SEQ'].strip('.'))
            # Germline positions
            result['d_germ_start'] = len(record['D_SEQ']) - len(record['D_SEQ'].lstrip('.'))
            result['d_germ_length'] = result['d_seq_length']

        return result

    @staticmethod
    def _parseJHit(record, db):
        """
        Parse J alignment information

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.
          db : database containing V and D alignment information.

        Returns:
          dict : database entries containing J call and alignment positions.
        """

        # J start position
        def _jstart():
            # J positions
            nb = [db['v_seq_start'],
                  db['v_seq_length'],
                  db['np1_length'],
                  db['d_seq_length'],
                  db['np2_length']]
            return sum(int(i) for i in nb if i)

        # Default return
        result = {'j_seq_start': None,
                  'j_seq_length': None,
                  'j_germ_start': None,
                  'j_germ_length': None}

        # Find J region
        if db['j_call']:
            # Query positions
            result['j_seq_start'] = _jstart()
            result['j_seq_length'] = len(record['J_SEQ'].strip('.'))
            # Germline positions
            result['j_germ_start'] = len(record['J_SEQ']) - len(record['J_SEQ'].lstrip('.'))
            result['j_germ_length'] = result['j_seq_length']

        return result

    @staticmethod
    def _assembleVDJ(record, db):
        """
        Build full length V(D)J sequence

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.
          db : database containing V and D alignment information.

        Returns:
          dict : database entries containing the full length V(D)J sequence.
        """
        segments = [record['V_SEQ'].strip('.') if db['v_call'] else '',
                    record['NP1_SEQ'] if db['np1_length'] else '',
                    record['D_SEQ'].strip('.') if db['d_call'] else '',
                    record['NP2_SEQ'] if db['np2_length'] else '',
                    record['J_SEQ'].strip('.') if db['j_call'] else '']

        return {'sequence_vdj': ''.join(segments)}

    @staticmethod
    def _parseScores(record):
        """
        Parse alignment scores

        Arguments:
          record : dictionary containing a single row from the iHMMune-Align ouptut.

        Returns:
          dict : database entries for alignment scores.
        """
        result = {}
        try:
            result['vdj_score'] = float(record['HMM_SCORE'])
        except (TypeError, ValueError):
            result['vdj_score'] = None

        return result

    def parseRecord(self, record):
        """
        Parses a single row from each IMTG file.

        Arguments:
          record : dictionary containing one row of iHMMune-Align file.

        Returns:
          dict : database entry for the row.
        """
        # Extract query ID and sequence
        query = record['SEQUENCE_ID']
        db = {'sequence_id': query,
              'sequence_input': str(self.sequences[query].seq)}

        # Check for valid alignment
        if not record['V_CALL'] or \
                record['V_CALL'].startswith('NA - ') or \
                record['V_CALL'].startswith('State path'):
            db['functional'] = None
            db['v_call'] = None
            db['d_call'] = None
            db['j_call'] = None
            return db

        # Parse record
        db.update(IHMMuneReader._parseFunctionality(record))
        db.update(IHMMuneReader._parseGenes(record))
        db.update(IHMMuneReader._parseNPHit(record))
        db.update(IHMMuneReader._parseVHit(record, db))
        db.update(IHMMuneReader._parseDHit(record, db))
        db.update(IHMMuneReader._parseJHit(record, db))
        db.update(IHMMuneReader._assembleVDJ(record, db))

        # Create IMGT-gapped sequence
        if 'v_call' in db and db['v_call'] and 'sequence_vdj' in db and db['sequence_vdj']:
            try:
                imgt_dict = gapV(db['sequence_vdj'],
                                 v_germ_start=db['v_germ_start_vdj'],
                                 v_germ_length=db['v_germ_length_vdj'],
                                 v_call=db['v_call'],
                                 references=self.references)
            except KeyError as e:
                imgt_dict = {'sequence_imgt': None,
                             'v_germ_start_imgt': None,
                             'v_germ_length_imgt': None}
                printWarning(e)
            db.update(imgt_dict)

        # Infer IMGT junction
        if ('j_call' in db and db['j_call']) and ('sequence_imgt' in db and db['sequence_imgt']):
            junc_dict = inferJunction(db['sequence_imgt'],
                                      j_germ_start=db['j_germ_start'],
                                      j_germ_length=db['j_germ_length'],
                                      j_call=db['j_call'],
                                      references=self.references,
                                      regions='default')
            db.update(junc_dict)

        # Overall alignment score
        db.update(IHMMuneReader._parseScores(record))

        # FWR and CDR regions
        rd = RegionDefinition(junction_length=db.get('junction_length', None),
                              amino_acid=False, definition='default')
        db.update(rd.getRegions(db.get('sequence_imgt', None)))
        
        return db

    def __iter__(self):
        """
        Iterator initializer.

        Returns:
          changeo.IO.IHMMuneReader
        """
        return self

    def __next__(self):
        """
        Next method.

        Returns:
          changeo.Receptor.Receptor : parsed IMGT/HighV-QUEST result as an Receptor (receptor=True) or dictionary (receptor=False).
        """
        # Get next set of records from dictionary readers
        try:
            record = None
            while not record:
                record = next(self.records)
        except StopIteration:
            raise StopIteration

        db = self.parseRecord(record)

        if self.receptor:
            return Receptor(db)
        else:
            return db


def readGermlines(references, asis=False, warn=False):
    """
    Parses germline repositories

    Arguments:
      references (list): list of strings specifying directories and/or files from which to read germline records.
      asis (bool): if True use sequence ID as record name and do not parse headers for allele names.
      warn (bool): print warning messages to standard error if True.

    Returns:
      dict: Dictionary of germlines in the form {allele: sequence}.
    """
    repo_files = []
    # Iterate over items passed to commandline
    for r in references:
        if os.path.isdir(r):
            # If directory, get fasta files from within
            repo_files.extend([os.path.join(r, f) for f in os.listdir(r) \
                          if getFileType(f) == 'fasta'])
        elif os.path.isfile(r) and getFileType(r) == 'fasta':
            # If file, make sure file is fasta
            repo_files.extend([r])

    # Catch instances where no valid fasta files were passed in
    if len(repo_files) < 1:
        printError('No valid germline fasta files (.fasta, .fna, .fa) were found at %s.' % ','.join(references))

    repo_dict = {}
    duplicates = []
    for file_name in repo_files:
        with open(file_name, 'r') as file_handle:
            germlines = SeqIO.parse(file_handle, 'fasta')
            for g in germlines:
                germ_key = getAllele(g.description, 'first') if not asis else g.id
                if germ_key not in repo_dict:
                    repo_dict[germ_key] = str(g.seq).upper()
                else:
                    duplicates.append(g.description)

    if warn and len(duplicates) > 0:
        w = indent('\n'.join(duplicates), ' '*9)
        printWarning('Duplicated germline allele names excluded from references:\n%s' % w)

    return repo_dict


def extractIMGT(imgt_output):
    """
    Extract necessary files from IMGT/HighV-QUEST results.

    Arguments:
      imgt_output : zipped file or unzipped folder output by IMGT/HighV-QUEST.

    Returns:
      tuple : (temporary directory handle, dictionary with names of extracted IMGT files).
    """
    # Map of IMGT file names
    imgt_names = ('1_Summary', '2_IMGT-gapped', '3_Nt-sequences', '6_Junction')
    imgt_keys = ('summary', 'gapped', 'ntseq', 'junction')

    # Open temporary directory and intialize return dictionary
    temp_dir = TemporaryDirectory()

    # Zip input
    if zipfile.is_zipfile(imgt_output):
        imgt_zip = zipfile.ZipFile(imgt_output, 'r')
        # Extract required files
        imgt_files = sorted([n for n in imgt_zip.namelist() \
                             if os.path.basename(n).startswith(imgt_names)])
        if len(imgt_files) < len(imgt_names):
            print("")
            printError('Missing necessary file(s) in IMGT output %s.' % imgt_output + ' Expecting:' + ', '.join(imgt_names))
            
        imgt_zip.extractall(temp_dir.name, imgt_files)
        # Define file dictionary
        imgt_dict = {k: os.path.join(temp_dir.name, f) for k, f in zip_longest(imgt_keys, imgt_files)}
    # Folder input
    elif os.path.isdir(imgt_output):
        folder_files = []
        for root, dirs, files in os.walk(imgt_output):
            folder_files.extend([os.path.join(os.path.abspath(root), f) for f in files])
        # Define file dictionary
        imgt_files = sorted([n for n in folder_files \
                             if os.path.basename(n).startswith(imgt_names)])
        if len(imgt_files) < len(imgt_names):
            print("")
            printError('Missing necessary file(s) in IMGT output %s.' % imgt_output + ' Expecting:' + ', '.join(imgt_names))
            
        imgt_dict = {k: f for k, f in zip_longest(imgt_keys, imgt_files)}
    # Tarball input
    elif tarfile.is_tarfile(imgt_output):
        imgt_tar = tarfile.open(imgt_output, 'r')
        # Extract required files
        imgt_files = sorted([n for n in imgt_tar.getnames() \
                             if os.path.basename(n).startswith(imgt_names)])
        if len(imgt_files) < len(imgt_names):
            print("")
            printError('Missing necessary file(s) in IMGT output %s.' % imgt_output + ' Expecting:' + ', '.join(imgt_names))
            
        imgt_tar.extractall(temp_dir.name, [imgt_tar.getmember(n) for n in imgt_files])
        # Define file dictionary
        imgt_dict = {k: os.path.join(temp_dir.name, f) for k, f in zip_longest(imgt_keys, imgt_files)}
    else:
        printError('Unsupported IGMT output file. Must be either a zipped file (.zip), LZMA compressed tarfile (.txz) or a folder.')

    # Check extraction for errors
    if len(imgt_dict) != len(imgt_names):
        printError('Extra files or missing necessary file IMGT output %s.' % imgt_output)

    return temp_dir, imgt_dict


def countDbFile(file):
    """
    Counts the records in database files

    Arguments:
      file : tab-delimited database file.

    Returns:
      int : count of records in the database file.
    """
    # Count records and check file
    try:
        with open(file, 'rt') as db_handle:
            db_records = csv.reader(db_handle, dialect='excel-tab')
            for i, __ in enumerate(db_records):  pass
        db_count = i
    except IOError:
        printError('File %s cannot be read.' % file)
    except:
        printError('File %s is invalid.' % file)
    else:
        if db_count == 0:  printError('File %s is empty.' % file)

    return db_count


def getDbFields(file, add=None, exclude=None, reader=TSVReader):
    """
    Get field names from a db file

    Arguments:
      file : db file to pull base fields from.
      add : fields to append to the field set.
      exclude : fields to exclude from the field set.
      reader : reader class.

    Returns:
        list : list of field names
    """
    try:
        with open(file, 'rt') as handle:
            fields = reader(handle).fields
    except IOError:
        printError('File %s cannot be read.' % file)
    except:
        printError('File %s is invalid.' % file)

    # Add extra fields
    if add is not None:
        if not isinstance(add, list):  add = [add]
        fields.extend([f for f in add if f not in fields])
    # Remove unwanted fields
    if exclude is not None:
        if not isinstance(exclude, list):  exclude = [exclude]
        fields = [f for f in fields if f not in exclude]

    return fields


def getFormatOperators(format):
    """
    Simple wrapper for fetching the set of operator classes for a data format

    Arguments:
      format (str): name of the data format.

    Returns:
      tuple: a tuple with the reader class, writer class, and schema definition class.
    """
    # Format options
    if format == 'changeo':
        reader = ChangeoReader
        writer = ChangeoWriter
        schema = ChangeoSchema
    elif format == 'changeo-aa':
        reader = ChangeoReader
        writer = ChangeoWriter
        schema = ChangeoSchemaAA
    elif format == 'airr':
        reader = AIRRReader
        writer = AIRRWriter
        schema = AIRRSchema
    elif format == 'airr-aa':
        reader = AIRRReader
        writer = AIRRWriter
        schema = AIRRSchemaAA
    else:
        raise ValueError

    return reader, writer, schema


def splitName(file):
    """
    Extract the extension from a file name

    Arguments:
      file (str): file name.

    Returns:
      tuple : tuple of the file directory, basename and extension.
    """
    directory, filename = os.path.split(file)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower().lstrip('.')

    return directory, basename, extension


def getOutputName(file, out_label=None, out_dir=None, out_name=None, out_type=None):
    """
    Creates and output filename from an existing filename

    Arguments:
      file : filename to base output file name on.
      out_label : text to be inserted before the file extension;
                  if None do not add a label.
      out_type : the file extension of the output file;
                 if None use input file extension.
      out_dir : the output directory;
                if None use directory of input file
      out_name : the short filename to use for the output file;
                 if None use input file short name.

    Returns:
      str: file name.
    """
    # Get filename components
    directory, basename, extension = splitName(file)

    # Define output directory
    if out_dir is None:
        out_dir = directory
    else:
        out_dir = os.path.abspath(out_dir)
        if not os.path.exists(out_dir):  os.mkdir(out_dir)
    # Define output file prefix
    if out_name is None:  out_name = basename
    # Define output file extension
    if out_type is None:  out_type = extension

    # Define output file name
    if out_label is None:
        out_file = os.path.join(out_dir, '%s.%s' % (out_name, out_type))
    else:
        out_file = os.path.join(out_dir, '%s_%s.%s' % (out_name, out_label, out_type))

    # Return file name
    return out_file


def getOutputHandle(file, out_label=None, out_dir=None, out_name=None, out_type=None):
    """
    Opens an output file handle

    Arguments:
      file : filename to base output file name on.
      out_label : text to be inserted before the file extension;
                  if None do not add a label.
      out_type : the file extension of the output file;
                 if None use input file extension.
      out_dir : the output directory;
                if None use directory of input file
      out_name : the short filename to use for the output file;
                 if None use input file short name.

    Returns:
      file : File handle
    """
    out_file = getOutputName(file, out_label=out_label, out_dir=out_dir,
                             out_name=out_name, out_type=out_type)

    # Open and return handle
    try:
        return open(out_file, mode='w')
    except:
        printError('File %s cannot be opened.' % out_file)


def checkFields(attributes, header, schema=AIRRSchema):
    """
    Checks that a file header contains a required set of Receptor attributes

    Arguments:
        attributes (list): list of Receptor attributes to check for.
        header (list): list of fields names in the file header.
        schema (object): schema object to convert field names to Receptor attributes.

    Returns:
        bool: True if all attributes mapping fields are found.

    Raises:
        LookupError:
    """
    if schema is None:  columns = attributes
    else:  columns = [schema.fromReceptor(f) for f in attributes]
    missing = [x for x in columns if x not in header]

    if len(missing) > 0:
        raise LookupError('Missing required fields: %s' % ', '.join(missing))

    return True


def yamlDict(file):
    """
    Returns a dictionary from a yaml file

    Arguments:
      file (str): simple yaml file with rows in the form 'argument: value'.

    Returns:
      dict: dictionary of key:value pairs in the file.
    """
    try:
        yaml_dict = dict(yaml.load(open(file, 'r'), Loader=yaml.FullLoader))
    except:
        printError('YAML file is invalid.')

    return yaml_dict



"""
Commandline interface
"""

# Info
__author__    = 'Jason Anthony Vander Heiden, Namita Gupta'
from changeo import __version__, __date__

# Imports
import os
import sys
import multiprocessing as mp
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, \
                     RawDescriptionHelpFormatter

# Changeo imports
from presto.IO import printWarning, printError
from changeo.Defaults import choices_format, default_format
from changeo.Receptor import AIRRSchema, ChangeoSchema


class CommonHelpFormatter(RawDescriptionHelpFormatter, ArgumentDefaultsHelpFormatter):
    """
    Custom argparse.HelpFormatter
    """
    # TODO:  add some sort of list formating for arguments with choices
    # TODO:  override argument position.
    # def __init__(self, prog, indent_increment=2, max_help_position=10, width=None):
    #    super(self.__class__, self).__init__(self, prog,
    #                                         indent_increment=indent_increment,
    #                                         max_help_position=max_help_position,
    #                                         width=width)

    # TODO:  remove multiple inheritance and clean up default value printing.
    # From ArgumentDefaultsHelpFormatter
    # def _get_help_string(self, action):
    #     help = action.help
    #     if '%(default)' not in action.help:
    #         if action.default is not SUPPRESS:
    #             defaulting_nargs = [OPTIONAL, ZERO_OR_MORE]
    #             if action.option_strings or action.nargs in defaulting_nargs:
    #                 help += ' (default: %(default)s)'
    #     return help
    pass


def getCommonArgParser(db_in=True, db_out=True, out_file=True, failed=True, log=True,
                       format=True, multiproc=False, add_help=True):
    """
    Defines an ArgumentParser object with common pRESTO arguments

    Arguments:
      db_in (bool): if True include tab delimited database input arguments.
      db_out (bool): if True include explicit output file name argument.
      out_file (bool): if True add explicit output file name arguments.
      failed (bool): if True include arguments for output of failed results.
      log (bool): if True include log arguments.
      format (bool): input and output type arguments.
      multiproc (bool): if True include multiprocessing arguments.

    Returns:
      argparse.ArgumentParser : an argument parser.
    """
    parser = ArgumentParser(formatter_class=CommonHelpFormatter, add_help=False)

    # Add help and version arguments
    if add_help:
        group_help = parser.add_argument_group('help')
        group_help.add_argument('--version', action='version',
                                version='%(prog)s:' + ' %s %s' %(__version__, __date__))
        group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')

    # Set standard group
    group = parser.add_argument_group('standard arguments')

    # Database arguments
    if db_in:
        group.add_argument('-d', nargs='+', action='store', dest='db_files', required=True,
                           help='A list of tab delimited database files.')
    if db_out:
        # Place holder for the future
        pass

    # Output filename
    if out_file:
        group.add_argument('-o', nargs='+', action='store', dest='out_files', default=None,
                           help='''Explicit output file name. Note, this argument cannot be used with 
                                 the --failed, --outdir, or --outname arguments. If unspecified, then
                                 the output filename will be based on the input filename(s).''')

    # Universal arguments
    group.add_argument('--outdir', action='store', dest='out_dir', default=None,
                       help='''Specify to changes the output directory to the location specified.
                            The input file directory is used if this is not specified.''')
    group.add_argument('--outname', action='store', dest='out_name', default=None,
                       help='''Changes the prefix of the successfully processed output file
                            to the string specified. May not be specified with multiple
                            input files.''')

    # Log arguments
    if log:
        group.add_argument('--log', action='store', dest='log_file', default=None,
                           help='''Specify to write verbose logging to a file. May not be
                                specified with multiple input files.''')

    # Failed result arguments
    if failed:
        group.add_argument('--failed', action='store_true', dest='failed',
                           help='''If specified create files containing records that
                                fail processing.''')
    # Format arguments
    if format:
        group.add_argument('--format', action='store', dest='format',
                           default=default_format, choices=choices_format,
                           help='''Output format. Also specifies the input format for tools accepting 
                                tab delimited AIRR Rearrangement or Change-O files.''')

    # Multiprocessing arguments
    if multiproc:
        group.add_argument('--nproc', action='store', dest='nproc', type=int, default=mp.cpu_count(),
                           help='''The number of simultaneous computational processes to execute
                                (CPU cores to utilized).''')

    return parser


def parseCommonArgs(args, in_arg=None, in_types=None, in_list=False):
    """
    Checks common arguments from getCommonArgParser and transforms output options to a dictionary

    Arguments: 
      args : Argument Namespace defined by ArgumentParser.parse_args.
      in_arg : String defining a non-standard input file argument to verify;
               by default 'db_files' and 'seq_files' are supported in that order.
      in_types : List of types (file extensions as strings) to allow for files in file_arg;
                 if None do not check type.
      in_list : if True allow multiple input files with the out_name and log arguments.
                    
    Returns:
      dict : Dictionary copy of args with output arguments embedded in the dictionary out_args
    """ 
    db_types = ['.tab', '.tsv', '.txt']
    seq_types = ['.fasta', 'fna', '.fa', '.fastq', '.fq']
    if in_types is not None:
        in_types = [f.lower for f in in_types]
    args_dict = args.__dict__.copy()
    
    # Count input files
    if 'seq_files' in args_dict:
        input_count = len(args_dict['seq_files']  or [])
        input_files = args_dict['seq_files']
    elif 'db_files' in args_dict:
        input_count = len(args_dict['db_files'] or [])
        input_files = args_dict['db_files']
    elif in_arg is not None and in_arg in args_dict:
        input_count = len(args_dict[in_arg] or [])
        input_files = args_dict[in_arg]
    else:
        printError('Cannot determine input file argument.')

    # Verify sequence files
    if 'seq_files' in args_dict and args_dict['seq_files']:
        for f in args_dict['seq_files']:
            if not os.path.isfile(f):
                printError('Sequence file %s does not exist.' % f)
            if os.path.splitext(f)[-1].lower() not in seq_types:
                printError('Sequence file %s is not a supported type. Must be one: %s.' \
                           % (f, ', '.join(seq_types)))

    # Verify database files
    if 'db_files' in args_dict and args_dict['db_files']:
        for f in args_dict['db_files']:
            if not os.path.isfile(f):
                printError('Database file %s does not exist.' % f)
            if os.path.splitext(f)[-1].lower() not in db_types:
                printError('Database file %s is not a supported type. Must be one: %s.' \
                           % (f, ', '.join(db_types)))

    # Verify non-standard input files
    if in_arg is not None and in_arg in args_dict and args_dict[in_arg]:
        files = args_dict[in_arg] if isinstance(args_dict[in_arg], list) else [args_dict[in_arg]]
        for f in files:
            if not os.path.exists(f):
                printError('Input %s does not exist.' % f)
            if in_types is not None and os.path.splitext(f)[-1].lower() not in in_types:
                printError('Input %s is not a supported type. Must be one: %s.' \
                         % (f, ', '.join(in_types)))

    # Verify output file arguments and exit if anything is hinky
    if args_dict.get('out_files', None) is not None \
            or args_dict.get('out_file', None) is not None:
        if args_dict.get('out_dir', None) is not None:
            printError('The -o argument may not be specified with the --outdir argument.')
        if args_dict.get('out_name', None) is not None:
            printError('The -o argument may not be specified with the --outname argument.')
        if args_dict.get('failed', False):
            printError('The -o argument may not be specified with the --failed argument.')
    if args_dict.get('out_files', None) is not None:
        if len(args_dict['out_files']) != input_count:
            printError('The -o argument requires one output file name per input file.')
        for f in args_dict['out_files']:
            if f in input_files:
                printError('Output files and input files cannot have the same names.')
        for f in args_dict['out_files']:
            if os.path.isfile(f):
                printWarning('Output file %s already exists and will be overwritten.' % f)
    if args_dict.get('out_file', None) is not None:
        if args_dict['out_file'] in input_files:
            printError('Output files and input files cannot have the same names.')
        if os.path.isfile(args_dict['out_file']):
            printWarning('Output file %s already exists and will be overwritten.' % args_dict['out_file'])

    # Exit if output names or log files are specified with multiple input files
    if args_dict.get('out_name', None) is not None \
            and input_count > 1 and not in_list:
        printError('The --outname argument may not be specified with multiple input files.')
    if args_dict.get('log_file', None) is not None \
            and input_count > 1 and not in_list:
        printError('The --log argument may not be specified with multiple input files.')
    
    # Verify output directory
    if 'out_dir' in args_dict and args_dict['out_dir']:
        if os.path.exists(args_dict['out_dir']) and not os.path.isdir(args_dict['out_dir']):
            printError('Path %s exists but it is not a directory.' % args_dict['out_dir'])

    # Redefine common output options as out_args dictionary
    out_args = ['log_file', 'out_dir', 'out_name', 'out_type', 'failed']
    args_dict['out_args'] = {k:args_dict.setdefault(k, None) for k in out_args}
    for k in out_args: del args_dict[k]
    
    return args_dict


def checkArgs(parser):
    """
    Checks that arguments have been provided and prints help if they have not.

    Arguments:
      parser : An argparse.ArgumentParser defining the commandline arguments.

    Returns:
      boolean : True if arguments are present. Prints help and exits if not.
    """
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    return True

def setDefaultFields(args, defaults, format='airr'):
    """
    Sets default field arguments by format

    Arguments:
      args (dict): parsed argument dictionary.
      defaults (dict): default variables to set with with keys as argument variables and values
                       as AIRR field names.
      format (str): one of 'changeo' or 'airr' which defines the file format.

    Returns:
      dict: modified input args.
    """
    if format == 'changeo':
        defaults = {k: ChangeoSchema.fromReceptor(AIRRSchema.toReceptor(v)) \
                    for k, v in defaults.items()}
    for f in defaults:
        if args[f] is None:  args[f] = defaults[f]

    return(args)



#!/usr/bin/env python3
"""
Assign V(D)J gene annotations
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import shutil
from argparse import ArgumentParser
from collections import OrderedDict
from pkg_resources import parse_version
from textwrap import dedent
from time import time
import re
import Bio
from Bio import SeqIO

# Presto imports
from presto.IO import printLog, printMessage, printError, printWarning
from changeo.Defaults import default_igblastn_exec, default_igblastp_exec, default_out_args
from changeo.Applications import runIgBLASTN, runIgBLASTP, getIgBLASTVersion
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from changeo.IO import getOutputName

# Defaults
choices_format = ('blast', 'airr')
choices_loci = ('ig', 'tr')
choices_organism = ('human', 'mouse', 'rabbit', 'rat', 'rhesus_monkey')
default_format = 'blast'
default_loci = 'ig'
default_organism = 'human'
default_igdata = '~/share/igblast'


def assignIgBLAST(seq_file, amino_acid=False, igdata=default_igdata, loci='ig', organism='human',
                  vdb=None, ddb=None, jdb=None, cdb=None, format=default_format,
                  igblast_exec=default_igblastn_exec, out_file=None,
                  out_args=default_out_args, nproc=None):
    """
    Performs clustering on sets of sequences

    Arguments:
      seq_file (str): the sample sequence file name.
      amino_acid : if True then run igblastp. igblastn is assumed if False.
      igdata (str): path to the IgBLAST database directory (IGDATA environment).
      loci (str): receptor type; one of 'ig' or 'tr'.
      organism (str): species name.
      vdb (str): name of a custom V reference in the database folder to use.
      ddb (str): name of a custom D reference in the database folder to use.
      jdb (str): name of a custom J reference in the database folder to use.
      cdb (str): name of a custom C reference in the database folder to use.
      format (str): output format. One of 'blast' or 'airr'.
      exec (str): the path to the igblastn executable.
      out_file (str): output file name. Automatically generated from the input file if None.
      out_args (dict): common output argument dictionary from parseCommonArgs.
      nproc (int): the number of processQueue processes;
              if None defaults to the number of CPUs.

    Returns:
      str: the output file name
    """
    # Check format argument
    try:
        out_type = {'blast': 'fmt7', 'airr': 'tsv'}[format]
    except KeyError:
        printError('Invalid output format %s.' % format)

    # Get IgBLAST version
    version = getIgBLASTVersion(exec=igblast_exec)
    if parse_version(version) < parse_version('1.6'):
        printError('IgBLAST version is %s and 1.6 or higher is required.' % version)
    if format == 'airr' and parse_version(version) < parse_version('1.9'):
        printError('IgBLAST version is %s and 1.9 or higher is required for AIRR format support.' % version)

    # Print parameter info
    log = OrderedDict()
    log['START'] = 'AssignGenes'
    log['COMMAND'] = 'igblast-aa' if amino_acid else 'igblast'
    log['VERSION'] = version
    log['FILE'] = os.path.basename(seq_file)
    log['ORGANISM'] = organism
    log['LOCI'] = loci
    log['NPROC'] = nproc
    printLog(log)

    # Open output writer
    if out_file is None:
        out_file = getOutputName(seq_file, out_label='igblast', out_dir=out_args['out_dir'],
                                 out_name=out_args['out_name'], out_type=out_type)
    # convert to FASTA if needed
    infile = open(seq_file, 'r')
    test = infile.read()[0]
    if test == "@":
        printMessage("Running conversion from FASTQ to FASTA")
        fasta_out_dir, filename = os.path.split(out_file)
        out_fasta_file = os.path.split(seq_file)[1]
        out_fasta_file = os.path.join(fasta_out_dir,'%s.fasta' % os.path.splitext(out_fasta_file)[0])
        with open(out_fasta_file, "w") as out_handle:
            records = SeqIO.parse(seq_file, 'fastq')
            if parse_version(Bio.__version__) >= parse_version('1.71'):
                # Biopython >= v1.71
                SeqIO.write(records, out_handle, format='fasta-2line')
            else:
                # Biopython < v1.71
                writer = SeqIO.FastaIO.FastaWriter(out_handle, wrap=None)
                writer.write_file(records)
        seq_file = out_fasta_file

    # Run IgBLAST clustering
    start_time = time()
    printMessage('Running IgBLAST', start_time=start_time, width=25)
    if not amino_acid:
        console_out = runIgBLASTN(seq_file, igdata, loci=loci, organism=organism,
                                  vdb=vdb, ddb=ddb, jdb=jdb, cdb=cdb, output=out_file,
                                  format=format, threads=nproc, exec=igblast_exec)
    else:
        console_out = runIgBLASTP(seq_file, igdata, loci=loci, organism=organism,
                                  vdb=vdb, output=out_file,
                                  threads=nproc, exec=igblast_exec)
    printMessage('Done', start_time=start_time, end=True, width=25)

    # Get number of processed sequences
    if (format == 'blast'):
        with open(out_file, 'rb') as f:
            f.seek(-2, os.SEEK_END)
            while f.read(1) != b'\n':
                f.seek(-2, os.SEEK_CUR)
            pass_info = f.readline().decode()
        num_seqs_match = re.search(r'(# BLAST processed )(\d+)( .*)', pass_info)
        num_sequences = num_seqs_match.group(2)
    else:
        f = open(out_file, 'rb')
        lines = 0
        buf_size = 1024 * 1024
        read_f = f.raw.read
        buf = read_f(buf_size)
        while buf:
            lines += buf.count(b'\n')
            buf = read_f(buf_size)
        num_sequences = lines - 1

    # Print log
    log = OrderedDict()
    log['PASS'] = num_sequences
    log['OUTPUT'] = os.path.basename(out_file)
    log['END'] = 'AssignGenes'
    printLog(log)

    return out_file


def getArgParser():
    """
    Defines the ArgumentParser

    Arguments:
    None

    Returns:
    an ArgumentParser object
    """
    # Define output file names and header fields
    fields = dedent(
             '''
             output files:
                 igblast
                    Reference alignment results from IgBLAST.
             ''')

    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', dest='command', metavar='',
                                       help='Assignment operation')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Parent parser
    parent_parser = getCommonArgParser(db_in=False, log=False, failed=False, format=False, multiproc=True)

    # Subparser to run igblastn
    parser_igblast = subparsers.add_parser('igblast', parents=[parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Executes igblastn.',
                                           description='Executes igblastn.')
    group_igblast = parser_igblast.add_argument_group('alignment arguments')
    group_igblast.add_argument('-s', nargs='+', action='store', dest='seq_files', required=True,
                               help='A list of FASTA files containing sequences to process.')
    group_igblast.add_argument('-b', action='store', dest='igdata', required=True,
                               help='IgBLAST database directory (IGDATA).')
    group_igblast.add_argument('--organism', action='store', dest='organism', default=default_organism,
                               choices=choices_organism, help='Organism name.')
    group_igblast.add_argument('--loci', action='store', dest='loci', default=default_loci,
                               choices=choices_loci, help='The receptor type.')
    group_igblast.add_argument('--vdb', action='store', dest='vdb', default=None,
                               help='''Name of the custom V reference in the IgBLAST database folder.
                                    If not specified, then a default database name with the form
                                    imgt_<organism>_<loci>_v will be used.''')
    group_igblast.add_argument('--ddb', action='store', dest='ddb', default=None,
                               help='''Name of the custom D reference in the IgBLAST database folder.
                                    If not specified, then a default database name with the form
                                    imgt_<organism>_<loci>_d will be used.''')
    group_igblast.add_argument('--jdb', action='store', dest='jdb', default=None,
                               help='''Name of the custom J reference in the IgBLAST database folder.
                                    If not specified, then a default database name with the form
                                    imgt_<organism>_<loci>_j will be used.''')
    group_igblast.add_argument('--cdb', action='store', dest='cdb', default=None,
                               help='''Name of the custom C reference in the IgBLAST database folder.
                                    If not specified, then a default database name with the form 
                                    imgt_<organism>_<loci>_c will be used. Note, this argument will be 
                                    ignored for IgBLAST versions below 1.18.0.''')
    group_igblast.add_argument('--format', action='store', dest='format', default=default_format,
                               choices=choices_format,
                               help='''Specify the output format. The "blast" will result in
                                    the IgBLAST "-outfmt 7 std qseq sseq btop" output format.
                                    Specifying "airr" will output the AIRR TSV format provided by
                                    the IgBLAST argument "-outfmt 19".''')
    group_igblast.add_argument('--exec', action='store', dest='igblast_exec',
                               default=default_igblastn_exec,
                               help='Path to the igblastn executable.')
    parser_igblast.set_defaults(func=assignIgBLAST, amino_acid=False)

    # Subparser to run igblastp
    parser_igblast_aa = subparsers.add_parser('igblast-aa', parents=[parent_parser],
                                           formatter_class=CommonHelpFormatter, add_help=False,
                                           help='Executes igblastp.',
                                           description='Executes igblastp.')
    group_igblast_aa = parser_igblast_aa.add_argument_group('alignment arguments')
    group_igblast_aa.add_argument('-s', nargs='+', action='store', dest='seq_files', required=True,
                                  help='A list of FASTA files containing sequences to process.')
    group_igblast_aa.add_argument('-b', action='store', dest='igdata', required=True,
                                  help='IgBLAST database directory (IGDATA).')
    group_igblast_aa.add_argument('--organism', action='store', dest='organism', default=default_organism,
                                  choices=choices_organism, help='Organism name.')
    group_igblast_aa.add_argument('--loci', action='store', dest='loci', default=default_loci,
                                  choices=choices_loci, help='The receptor type.')
    group_igblast_aa.add_argument('--vdb', action='store', dest='vdb', default=None,
                                  help='''Name of the custom V reference in the IgBLAST database folder.
                                       If not specified, then a default database name with the form
                                       imgt_aa_<organism>_<loci>_v will be used.''')
    group_igblast_aa.add_argument('--exec', action='store', dest='igblast_exec',
                                  default=default_igblastp_exec,
                                  help='Path to the igblastp executable.')
    parser_igblast_aa.set_defaults(func=assignIgBLAST, amino_acid=True, ddb=None, jdb=None, format='blast')


    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Check if a valid clustering executable was specified
    if not shutil.which(args_dict['igblast_exec']):
        parser.error('%s executable not found' % args_dict['igblast_exec'])

    # Clean arguments dictionary
    del args_dict['seq_files']
    if 'out_files' in args_dict: del args_dict['out_files']
    del args_dict['func']
    del args_dict['command']

    # Call main function for each input file
    for i, f in enumerate(args.__dict__['seq_files']):
        args_dict['seq_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        args.func(**args_dict)



#!/usr/bin/env python3
"""
Reconstructs germline sequences from alignment data
"""

# Info
__author__ = 'Namita Gupta, Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
from argparse import ArgumentParser
from collections import OrderedDict
from itertools import groupby
from textwrap import dedent
from time import time

# Presto and change imports
from presto.Defaults import default_out_args
from presto.IO import printLog, printMessage, printProgress, printError, printWarning
from changeo.Defaults import default_v_field, default_d_field, default_j_field, default_clone_field, \
                             default_seq_field, default_format
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs, \
                                setDefaultFields
from changeo.Gene import buildGermline, buildClonalGermline
from changeo.IO import countDbFile, getDbFields, getFormatOperators, getOutputHandle, readGermlines, \
                       checkFields

# Defaults
default_germ_types = ['dmask']


def createGermlines(db_file, references, seq_field=default_seq_field, v_field=default_v_field,
                    d_field=default_d_field, j_field=default_j_field,
                    cloned=False, clone_field=default_clone_field, germ_types=default_germ_types,
                    format=default_format, out_file=None, out_args=default_out_args):
    """
    Write germline sequences to tab-delimited database file

    Arguments:
      db_file : input tab-delimited database file.
      references : folders and/or files containing germline repertoire data in FASTA format.
      seq_field : field in which to look for sequence.
      v_field : field in which to look for V call.
      d_field : field in which to look for D call.
      j_field : field in which to look for J call.
      cloned : if True build germlines by clone, otherwise build individual germlines.
      clone_field : field containing clone identifiers; ignored if cloned=False.
      germ_types : list of germline sequence types to be output from the set of 'full', 'dmask', 'vonly', 'regions'
      format : input and output format.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : arguments for output preferences.

    Returns:
      dict: names of the 'pass' and 'fail' output files.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'CreateGermlines'
    log['FILE'] = os.path.basename(db_file)
    log['GERM_TYPES'] = ','.join(germ_types)
    log['SEQ_FIELD'] = seq_field
    log['V_FIELD'] = v_field
    log['D_FIELD'] = d_field
    log['J_FIELD'] = j_field
    log['CLONED'] = cloned
    if cloned:  log['CLONE_FIELD'] = clone_field
    printLog(log)

    # Define format operators
    try:
        reader, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s' % format)
    out_args['out_type'] = schema.out_type

    # TODO: this won't work for AIRR necessarily
    # Define output germline fields
    germline_fields = OrderedDict()
    seq_type = seq_field.split('_')[-1]
    if 'full' in germ_types:  germline_fields['full'] = 'germline_' + seq_type
    if 'dmask' in germ_types:  germline_fields['dmask'] = 'germline_' + seq_type + '_d_mask'
    if 'vonly' in germ_types:  germline_fields['vonly'] = 'germline_' + seq_type + '_v_region'
    if 'regions' in germ_types:  germline_fields['regions'] = 'germline_regions'
    if cloned:
        germline_fields['v'] = 'germline_v_call'
        germline_fields['d'] = 'germline_d_call'
        germline_fields['j'] = 'germline_j_call'
    out_fields = getDbFields(db_file,
                             add=[schema.fromReceptor(f) for f in germline_fields.values()],
                             reader=reader)

    # Get repertoire and open Db reader
    reference_dict = readGermlines(references)
    db_handle = open(db_file, 'rt')
    db_iter = reader(db_handle)

    # Check for required columns
    try:
        required = ['v_germ_start_imgt', 'd_germ_start', 'j_germ_start',
                    'np1_length', 'np2_length']
        checkFields(required, db_iter.fields, schema=schema)
    except LookupError as e:
        printError(e)

    # Check for IMGT-gaps in germlines
    if all('...' not in x for x in reference_dict.values()):
        printWarning('Germline reference sequences do not appear to contain IMGT-numbering spacers. Results may be incorrect.')

    # Count input
    total_count = countDbFile(db_file)

    # Check for existence of fields
    for f in [v_field, d_field, j_field, seq_field]:
        if f not in db_iter.fields:
            printError('%s field does not exist in input database file.' % f)

    # Translate to Receptor attribute names
    v_field = schema.toReceptor(v_field)
    d_field = schema.toReceptor(d_field)
    j_field = schema.toReceptor(j_field)
    seq_field = schema.toReceptor(seq_field)
    clone_field = schema.toReceptor(clone_field)

    # Define Receptor iterator
    if cloned:
        start_time = time()
        printMessage('Sorting by clone', start_time=start_time, width=20)
        sorted_records = sorted(db_iter, key=lambda x: x.getField(clone_field))
        printMessage('Done', start_time=start_time, end=True, width=20)
        receptor_iter = groupby(sorted_records, lambda x: x.getField(clone_field))
    else:
        receptor_iter = ((x.sequence_id, [x]) for x in db_iter)

    # Define log handle
    if out_args['log_file'] is None:
        log_handle = None
    else:
        log_handle = open(out_args['log_file'], 'w')

    # Initialize handles, writers and counters
    pass_handle, pass_writer = None, None
    fail_handle, fail_writer = None, None
    rec_count, pass_count, fail_count = 0, 0, 0
    start_time = time()

    # Iterate over rows
    for key, records in receptor_iter:
        # Print progress
        printProgress(rec_count, total_count, 0.05, start_time=start_time)

        # Define iteration variables
        records = list(records)
        rec_log = OrderedDict([('ID', key)])
        rec_count += len(records)

        # Build germline for records
        if len(records) == 1:
            germ_log, germlines, genes = buildGermline(records[0], reference_dict, seq_field=seq_field, v_field=v_field,
                                                       d_field=d_field, j_field=j_field)
        else:
            germ_log, germlines, genes = buildClonalGermline(records, reference_dict, seq_field=seq_field, v_field=v_field,
                                                             d_field=d_field, j_field=j_field)
        rec_log.update(germ_log)

        # Write row to pass or fail file
        if germlines is not None:
            pass_count += len(records)

            # Add germlines to Receptor record
            annotations = {}
            if 'full' in germ_types:  annotations[germline_fields['full']] = germlines['full']
            if 'dmask' in germ_types:  annotations[germline_fields['dmask']] = germlines['dmask']
            if 'vonly' in germ_types:  annotations[germline_fields['vonly']] = germlines['vonly']
            if 'regions' in germ_types:  annotations[germline_fields['regions']] = germlines['regions']
            if cloned:
                annotations[germline_fields['v']] = genes['v']
                annotations[germline_fields['d']] = genes['d']
                annotations[germline_fields['j']] = genes['j']

            # Write records
            try:
                for r in records:
                    r.setDict(annotations)
                    pass_writer.writeReceptor(r)
            except AttributeError:
                # Create output file handle and writer
                if out_file is not None:
                    pass_handle = open(out_file, 'w')
                else:
                    pass_handle = getOutputHandle(db_file,
                                                  out_label='germ-pass',
                                                  out_dir=out_args['out_dir'],
                                                  out_name=out_args['out_name'],
                                                  out_type=out_args['out_type'])
                pass_writer = writer(pass_handle, fields=out_fields)
                for r in records:
                    r.setDict(annotations)
                    pass_writer.writeReceptor(r)
        else:
            fail_count += len(records)
            if out_args['failed']:
                try:
                    fail_writer.writeReceptor(records)
                except AttributeError:
                    fail_handle = getOutputHandle(db_file,
                                                  out_label='germ-fail',
                                                  out_dir=out_args['out_dir'],
                                                  out_name=out_args['out_name'],
                                                  out_type=out_args['out_type'])
                    fail_writer = writer(fail_handle, fields=out_fields)
                    fail_writer.writeReceptor(records)

        # Write log
        printLog(rec_log, handle=log_handle)

    # Print log
    printProgress(rec_count, total_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
    log['RECORDS'] = rec_count
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'CreateGermlines'
    printLog(log)

    # Close file handles
    db_handle.close()
    output = {'pass': None, 'fail': None}
    if pass_handle is not None:
        output['pass'] = pass_handle.name
        pass_handle.close()
    if fail_handle is not None:
        output['fail'] = fail_handle.name
        fail_handle.close()
    if log_handle is not None:
        log_handle.close()

    return output


def getArgParser():
    """
    Defines the ArgumentParser

    Arguments:
    None

    Returns:
    an ArgumentParser object
    """
    # Define input and output field help message
    fields = dedent(
             '''
             output files:
                 germ-pass
                    database with assigned germline sequences.
                 germ-fail
                    database with records failing germline assignment.

             required fields:
                 sequence_id, sequence_alignment, v_call, d_call, j_call, 
                 v_sequence_start, v_sequence_end, v_germline_start, v_germline_end,
                 d_sequence_start, d_sequence_end, d_germline_start, d_germline_end,
                 j_sequence_start, j_sequence_end, j_germline_start, j_germline_end,
                 np1_length, np2_length

             optional fields:
                 n1_length, n2_length, p3v_length, p5d_length, p3d_length, p5j_length,
                 clone_id

             output fields:
                 germline_v_call, germline_d_call, germline_j_call,
                 germline_alignment, germline_alignment_d_mask, 
                 germline_alignment_v_region, germline_regions, 
              ''')
    # Define argument parser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            parents=[getCommonArgParser(format=True)],
                            formatter_class=CommonHelpFormatter, add_help=False)

    # Germlines arguments
    group = parser.add_argument_group('germline construction arguments')
    group.add_argument('-r', nargs='+', action='store', dest='references', required=True,
                        help='''List of folders and/or fasta files (with .fasta, .fna or .fa
                         extension) with germline sequences. When using the default
                         Change-O sequence and coordinate fields, these reference sequences 
                         must contain IMGT-numbering spacers (gaps) in the V segment. 
                         Alternative numbering schemes, or no numbering, may work for alternative 
                         sequence and coordinate definitions that define a valid alignment, but 
                         a warning will be issued.''')
    group.add_argument('-g', action='store', dest='germ_types', default=default_germ_types,
                        nargs='+', choices=('full', 'dmask', 'vonly', 'regions'),
                        help='''Specify type(s) of germlines to include full germline,
                             germline with D segment masked, or germline for V segment only.''')
    group.add_argument('--cloned', action='store_true', dest='cloned',
                        help='''Specify to create only one germline per clone. Note, if allele 
                             calls are ambiguous within a clonal group, this will place the 
                             germline call used for the entire clone within the
                             germline_v_call, germline_d_call and germline_j_call fields.''')
    group.add_argument('--sf', action='store', dest='seq_field', default=None,
                        help='''Field containing the aligned sequence.
                             Defaults to sequence_alignment (airr) or SEQUENCE_IMGT (changeo).''')
    group.add_argument('--vf', action='store', dest='v_field', default=None,
                        help='''Field containing the germline V segment call.
                             Defaults to v_call (airr) or V_CALL (changeo).''')
    group.add_argument('--df', action='store', dest='d_field', default=None,
                        help='''Field containing the germline D segment call.
                             Defaults to d_call (airr) or D_CALL (changeo).''')
    group.add_argument('--jf', action='store', dest='j_field', default=None,
                        help='''Field containing the germline J segment call.
                             Defaults to j_call (airr) or J_CALL (changeo).''')
    group.add_argument('--cf', action='store', dest='clone_field', default=None,
                        help='''Field containing clone identifiers. 
                             Ignored if --cloned is not also specified.
                             Defaults to clone_id (airr) or CLONE (changeo).''')

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main
    """

    # Parse command line arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Set default fields
    default_fields = {'seq_field': default_seq_field,
                      'v_field': default_v_field,
                      'd_field': default_d_field,
                      'j_field': default_j_field,
                      'clone_field': default_clone_field}
    args_dict = setDefaultFields(args_dict, default_fields, format=args_dict['format'])

    # Check that reference files exist
    for f in args_dict['references']:
        if not os.path.exists(f):
            parser.error('Germline reference file or folder %s does not exist.' % f)

    # Clean arguments dictionary
    del args_dict['db_files']
    if 'out_files' in args_dict: del args_dict['out_files']

    # Call main function for each input file
    for i, f in enumerate(args.__dict__['db_files']):
        args_dict['db_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        createGermlines(**args_dict)



"""
Unit tests for MakeDb
"""
# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import os
import sys
import time
import unittest
from Bio import SeqIO

# Presto and changeo imports
from changeo.IO import getDbFields, extractIMGT, readGermlines, ChangeoReader, \
                       IgBLASTReader, IgBLASTReaderAA, IHMMuneReader, IMGTReader

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')

# Import script
sys.path.append(os.path.join(test_path, os.pardir, 'bin'))
import MakeDb


class Test_MakeDb(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)

        # Germline files
        self.repo_ig = '/usr/local/share/germlines/imgt/human/vdj'
        self.repo_ig_aa = '/usr/local/share/germlines/imgt/human/vdj_aa'
        # Read files
        self.reads_ig = os.path.join(data_path, 'reads_ig.fasta')
        self.reads_ig_aa = os.path.join(data_path, 'reads_ig_aa.fasta')
        # IMGT output
        self.imgt_ig = os.path.join(data_path, 'imgt_ig.txz')
        # IgBLAST output
        self.igblast_ig = os.path.join(data_path, 'igblast1.7_ig.fmt7')
        self.igblast_ig_aa = os.path.join(data_path, 'igblast1.14_ig_aa.fmt7')
        # iHMMune-Align output
        self.ihmmune_ig = os.path.join(data_path, 'ihmmune_ig.csv')
        # Change-O files
        self.db_ig = os.path.join(data_path, 'imgt_ig_db-pass.tsv')

        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print("<- %s() %.3f" % (self._testMethodName, t))

    @unittest.skip("-> ChangeoReader() skipped\n")
    def test_getDbFields(self):
        # Get fields
        x = getDbFields(self.db_ig)
        print(x)

        # Add fields
        x = getDbFields(self.db_ig, add=['A', 'B', 'C'])
        print(x)

        # Exclude fields
        x = getDbFields(self.db_ig, exclude=['V_SCORE', 'V_IDENTITY', 'J_SCORE', 'J_IDENTITY'])
        print(x)

        # Add and exclude fields
        x = getDbFields(self.db_ig, add=['A', 'B', 'C'],
                        exclude=['V_SCORE', 'V_IDENTITY', 'J_SCORE', 'J_IDENTITY'])
        print(x)

        self.fail('TODO')

    @unittest.skip("-> ChangeoReader() skipped\n")
    def test_ChangeoReader(self):
        # Parse
        with open(self.db_ig, 'r') as f:
            result = ChangeoReader(f)
            for x in result: print(x.toDict())

        self.fail('TODO')

    @unittest.skip("-> IMGTReader() skipped\n")
    def test_IMGTReader(self):
        # Extract IMGT files
        temp_dir, files = extractIMGT(self.imgt_ig)

        # Parse
        with open(files['summary'], 'r') as summary, \
                open(files['gapped'], 'r') as gapped, \
                open(files['ntseq'], 'r') as ntseq, \
                open(files['junction'], 'r') as junction:
            result = IMGTReader(summary, gapped, ntseq, junction, receptor=False)
            for x in result: print(x)

        # Remove IMGT temporary directory
        temp_dir.cleanup()

        self.fail('TODO')

    @unittest.skip("-> IgBLASTReader() skipped\n")
    def test_IgBLASTReader(self):
        # Load germlines and sequences
        seq_dict = MakeDb.getSeqDict(self.reads_ig_aa)
        repo_dict = readGermlines([self.repo_ig])

        # Parse
        with open(self.igblast_ig, 'r') as f:
            result = IgBLASTReader(f, seq_dict, repo_dict, receptor=False)
            for x in result: print(x)

        self.fail('TODO')

    @unittest.skip("-> IgBLASTAAReader() skipped\n")
    def test_IgBLASTAAReader(self):
        # Load germlines and sequences
        seq_dict = MakeDb.getSeqDict(self.reads_ig_aa)
        repo_dict = readGermlines([self.repo_ig_aa])

        # Parse
        with open(self.igblast_ig_aa, 'r') as f:
            result = IgBLASTReaderAA(f, seq_dict, repo_dict, receptor=False)
            for x in result: print(x)

        # self.fail('TODO')

    @unittest.skip("-> IHMMReader() skipped\n")
    def test_IHMMReader(self):
        # Load germlines and sequences
        seq_dict = MakeDb.getSeqDict(self.reads_ig)
        repo_dict = readGermlines([self.repo_ig])

        # Parse
        with open(self.ihmmune_ig, 'r') as f:
            result = IHMMuneReader(f, seq_dict, repo_dict, receptor=False)
            for x in result: print(x)

        self.fail('TODO')

if __name__ == '__main__':
    unittest.main()


"""
Alignment manipulation
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import re
from Bio.Seq import Seq

# Presto and changeo imports
from changeo.Gene import getVAllele, getJAllele

# Load regions
# import yaml
# from pkg_resources import resource_stream
# with resource_stream(__name__, 'data/regions.yaml') as f:
#     imgt_regions = yaml.load(f, Loader=yaml.FullLoader)
imgt_regions = {'default': {'fwr1': 1,
                            'cdr1': 27,
                            'fwr2': 39,
                            'cdr2': 56,
                            'fwr3': 66,
                            'cdr3': 105},
                'rhesus-igl': {'fwr1': 1,
                               'cdr1': 28,
                               'fwr2': 40,
                               'cdr2': 59,
                               'fwr3': 69,
                               'cdr3': 108}}


class RegionDefinition:
    """
    FWR and CDR region boundary definitions
    """
    def __init__(self, junction_length, amino_acid=False, definition='default'):
        """
        Initializer

        Arguments:
          junction_length (int): length of the junction region. If None then CDR3 end and FWR4 start/end are undefined.
          definition (str): region definition entry in the data/regions.yaml file to use.
          amino_acid (bool): if True define boundaries in amino acid space, otherwise use nucleotide positions.

        Returns:
          changeo.Alignment.RegionDefinition
        """
        self.junction_length = junction_length
        self.amino_acid = amino_acid
        self.definition = definition
        pos_mod = 1 if amino_acid else 3

        # Define regions
        regions = {k: (int(v) - 1) * pos_mod for k, v in imgt_regions[definition].items()}

        # Assign positions
        if junction_length is not None:
            fwr4_start = max(regions['cdr3'], regions['cdr3'] - (2 * pos_mod) + junction_length) \
                if junction_length is not None else None
            junction_end = fwr4_start + (1 * pos_mod)
        else:
            fwr4_start = None
            junction_end = None
        self.positions = {'fwr1': [regions['fwr1'], regions['cdr1']],
                          'cdr1': [regions['cdr1'], regions['fwr2']],
                          'fwr2': [regions['fwr2'], regions['cdr2']],
                          'cdr2': [regions['cdr2'], regions['fwr3']],
                          'fwr3': [regions['fwr3'], regions['cdr3']],
                          'cdr3': [regions['cdr3'], fwr4_start],
                          'fwr4': [fwr4_start, None],
                          'junction': [regions['cdr3'] - (1 * pos_mod), junction_end]}

    def getRegions(self, seq):
        """
        Return IMGT defined FWR and CDR regions

        Arguments:
          seq : IMGT-gapped sequence.

        Returns:
          dict : dictionary of FWR and CDR sequences.
        """
        regions = {'fwr1_imgt': None,
                   'fwr2_imgt': None,
                   'fwr3_imgt': None,
                   'fwr4_imgt': None,
                   'cdr1_imgt': None,
                   'cdr2_imgt': None,
                   'cdr3_imgt': None}
        try:
            seq_len = len(seq)
            regions['fwr1_imgt'] = seq[self.positions['fwr1'][0]:min(self.positions['fwr1'][1], seq_len)]
        except (KeyError, IndexError, TypeError):
            return regions

        try:
            regions['cdr1_imgt'] = seq[self.positions['cdr1'][0]:min(self.positions['cdr1'][1], seq_len)]
        except (IndexError):
            return regions

        try:
            regions['fwr2_imgt'] = seq[self.positions['fwr2'][0]:min(self.positions['fwr2'][1], seq_len)]
        except (IndexError):
            return regions

        try:
            regions['cdr2_imgt'] = seq[self.positions['cdr2'][0]:min(self.positions['cdr2'][1], seq_len)]
        except (IndexError):
            return regions

        try:
            regions['fwr3_imgt'] = seq[self.positions['fwr3'][0]:min(self.positions['fwr3'][1], seq_len)]
        except (IndexError):
            return regions

        try:
            regions['cdr3_imgt'] = seq[self.positions['cdr3'][0]:min(self.positions['cdr3'][1], seq_len)]
            regions['fwr4_imgt'] = seq[self.positions['fwr4'][0]:]
        except (KeyError, IndexError, TypeError):
            return regions

        return regions


def decodeBTOP(btop):
    """
    Parse a BTOP string into a list of tuples in CIGAR annotation.

    Arguments:
      btop : BTOP string.

    Returns:
      list : tuples of (operation, length) for each operation in the BTOP string using CIGAR annotation.
    """
    # Determine chunk type and length
    def _recode(m):
        if m.isdigit():  return ('=', int(m))
        elif m[0] == '-':  return ('I', len(m) // 2)
        elif m[1] == '-':  return ('D', len(m) // 2)
        else:  return ('X', len(m) // 2)

    # Split BTOP string into sections
    btop_split = re.sub(r'(\d+|[-A-Z]{2})', r'\1;', btop)
    # Parse each chunk of encoding
    matches = re.finditer(r'(\d+)|([A-Z]{2};)+|(-[A-Z];)+|([A-Z]-;)+', btop_split)

    return [_recode(m.group().replace(';', '')) for m in matches]


def decodeCIGAR(cigar):
    """
    Parse a CIGAR string into a list of tuples.

    Arguments:
      cigar : CIGAR string.

    Returns:
      list : tuples of (operation, length) for each operation in the CIGAR string.
    """
    matches = re.findall(r'(\d+)([A-Z])', cigar)

    return [(m[1], int(m[0])) for m in matches]


def encodeCIGAR(alignment):
    """
    Encodes a list of tuple with alignment information into a CIGAR string.

    Arguments:
      tuple : tuples of (type, length) for each alignment operation.

    Returns:
      str : CIGAR string.
    """
    return ''.join(['%i%s' % (x, s) for s, x in alignment])


def padAlignment(alignment, q_start, r_start):
    """
    Pads the start of an alignment based on query and reference positions.

    Arguments:
      alignment : tuples of (operation, length) for each alignment operation.
      q_start : query (input) start position (0-based)
      r_start : reference (subject) start position (0-based)

    Returns:
      list : updated list of tuples of (operation, length) for the alignment.
    """
    # Copy list to avoid weirdness
    result = alignment[:]

    # Add query deletions
    if result [0][0] == 'S':
        result[0] = ('S', result[0][1] + q_start)
    elif q_start > 0:
        result.insert(0, ('S', q_start))

    # Add reference padding if present
    if result[0][0] == 'N':
        result[0] = ('N', result[0][1] + r_start)
    elif result [0][0] == 'S' and result[1][0] == 'N':
        result[1] = ('N', result[1][1] + r_start)
    elif result[0][0] == 'S' and r_start > 0:
        result.insert(1, ('N', r_start))
    elif r_start > 0:
        result.insert(0, ('N', r_start))

    return result


def alignmentPositions(alignment):
    """
    Extracts start position and length from an alignment

    Arguments:
      alignment : tuples of (operation, length) for each alignment operation.

    Returns:
      dict : query (q) and reference (r) start (0-based) and length information with keys
             {q_start, q_length, r_start, r_length}.
    """
    # Return object
    result = {'q_start': 0,
              'q_length': 0,
              'r_start': 0,
              'r_length': 0}

    # Query start
    if alignment[0][0] == 'S':
        result['q_start'] = alignment[0][1]

    # Reference start
    if alignment[0][0] == 'N':
        result['r_start'] = alignment[0][1]
    elif alignment[0][0] == 'S' and alignment[1][0] == 'N':
        result['r_start'] = alignment[1][1]

    # Reference length
    for x, i in alignment:
        if x in ('M', '=', 'X'):
            result['r_length'] += i
            result['q_length'] += i
        elif x == 'D':
            result['r_length'] += i
        elif x == 'I':
            result['q_length'] += i

    return result


def gapV(seq, v_germ_start, v_germ_length, v_call, references, asis_calls=False):
    """
    Construction IMGT-gapped V segment sequences.

    Arguments:
      seq (str): V(D)J sequence alignment (SEQUENCE_VDJ).
      v_germ_start (int): start position V segment alignment in the germline (V_GERM_START_VDJ, 1-based).
      v_germ_length (int): length of the V segment alignment against the germline (V_GERM_LENGTH_VDJ, 1-based).
      v_call (str): V segment allele assignment (V_CALL).
      references (dict): dictionary of IMGT-gapped reference sequences.
      asis_calls (bool): if True do not parse v_call for allele names and just split by comma.

    Returns:
      dict: dictionary containing IMGT-gapped query sequences and germline positions.

    Raises:
      KeyError: raised if the v_call is not found in the reference dictionary.
    """
    # Initialize return object
    imgt_dict = {'sequence_imgt': None,
                 'v_germ_start_imgt': None,
                 'v_germ_length_imgt': None}

    # Initialize imgt gapped sequence
    seq_imgt = '.' * (int(v_germ_start) - 1) + seq

    # Extract first V call
    if not asis_calls:
        vgene = getVAllele(v_call, action='first')
    else:
        vgene = v_call.split(',')[0]

    # Find gapped germline V segment
    try:
    #if vgene in references:
        vgap = references[vgene]
        # Iterate over gaps in the germline segment
        gaps = re.finditer(r'\.', vgap)
        gapcount = int(v_germ_start) - 1
        for gap in gaps:
            i = gap.start()
            # Break if gap begins after V region
            if i >= v_germ_length + gapcount:
                break
            # Insert gap into IMGT sequence
            seq_imgt = seq_imgt[:i] + '.' + seq_imgt[i:]
            # Update gap counter
            gapcount += 1

        imgt_dict['sequence_imgt'] = seq_imgt
        # Update IMGT positioning information for V
        imgt_dict['v_germ_start_imgt'] = 1
        imgt_dict['v_germ_length_imgt'] = v_germ_length + gapcount
    except KeyError as e:
        raise KeyError('%s was not found in the germline repository.' % vgene)
    #else:
    #    printWarning('%s was not found in the germline repository. IMGT-gapped sequence cannot be determined.' % vgene)

    return imgt_dict


def inferJunction(seq, j_germ_start, j_germ_length, j_call, references, asis_calls=False, regions='default'):
    """
    Identify junction region by IMGT definition.

    Arguments:
      seq (str): IMGT-gapped V(D)J sequence alignment (SEQUENCE_IMGT).
      j_germ_start (int): start position J segment alignment in the germline (J_GERM_START, 1-based).
      j_germ_length (int): length of the J segment alignment against the germline (J_GERM_LENGTH).
      j_call (str): J segment allele assignment (J_CALL).
      references (dict): dictionary of IMGT-gapped reference sequences.
      asis_calls (bool): if True do not parse V_CALL for allele names and just split by comma.
      regions (str): name of the IMGT FWR/CDR region definitions to use.

    Returns:
      dict : dictionary containing junction sequence, translation and length.
    """
    junc_dict = {'junction': None,
                 'junction_aa': None,
                 'junction_length': None}

    # Find germline J segment
    if not asis_calls:
        jgene = getJAllele(j_call, action='first')
    else:
        jgene = j_call.split(',')[0]
    jgerm = references.get(jgene, None)

    if jgerm is not None:
        # Look for (F|W)GXG amino acid motif in germline nucleotide sequence
        motif = re.search(r'T(TT|TC|GG)GG[ACGT]{4}GG[AGCT]', jgerm)

        # Define junction end position
        seq_len = len(seq)
        if motif:
            j_start = seq_len - j_germ_length
            motif_pos = max(motif.start() - j_germ_start + 1, -1)
            junc_end = j_start + motif_pos + 3
        else:
            junc_end = seq_len

        # Extract junction
        rd = RegionDefinition(None, amino_acid=False, definition=regions)
        junc_start = rd.positions['junction'][0]
        junc_dict['junction'] = seq[junc_start:junc_end]
        junc_len = len(junc_dict['junction'])
        junc_dict['junction_length'] = junc_len

        # Translation
        junc_tmp = junc_dict['junction'].replace('-', 'N').replace('.', 'N')
        if junc_len % 3 > 0:  junc_tmp = junc_tmp[:junc_len - junc_len % 3]
        junc_dict['junction_aa'] = str(Seq(junc_tmp).translate())

    return junc_dict


def getRegions(seq, junction_length):
    """
    Identify FWR and CDR regions by IMGT definition.

    Arguments:
      seq : IMGT-gapped sequence.
      junction_length : length of the junction region in nucleotides.

    Returns:
      dict : dictionary of FWR and CDR sequences.
    """
    region_dict = {'fwr1_imgt': None,
                   'fwr2_imgt': None,
                   'fwr3_imgt': None,
                   'fwr4_imgt': None,
                   'cdr1_imgt': None,
                   'cdr2_imgt': None,
                   'cdr3_imgt': None}
    try:
        seq_len = len(seq)
        region_dict['fwr1_imgt'] = seq[0:min(78, seq_len)]
    except (KeyError, IndexError, TypeError):
        return region_dict

    try: region_dict['cdr1_imgt'] = seq[78:min(114, seq_len)]
    except (IndexError): return region_dict

    try: region_dict['fwr2_imgt'] = seq[114:min(165, seq_len)]
    except (IndexError): return region_dict

    try: region_dict['cdr2_imgt'] = seq[165:min(195, seq_len)]
    except (IndexError): return region_dict

    try: region_dict['fwr3_imgt'] = seq[195:min(312, seq_len)]
    except (IndexError): return region_dict

    try:
        # CDR3
        cdr3_end = 306 + junction_length
        region_dict['cdr3_imgt'] = seq[312:cdr3_end]
        # FWR4
        region_dict['fwr4_imgt'] = seq[cdr3_end:]
    except (KeyError, IndexError, TypeError):
        return region_dict

    return region_dict
